# -*- coding: utf-8 -*-
"""


Cache build requirements:

    * Caching mechanism should inspect 'inputs' (collect file hashes) to determine if any content has
      changed.  If input varies, then command should be re-run.
    * Command (decorated function) should be generally unaware of all other details of build process,
      and it should *ONLY* be able to see files listed in "inputs"
    * Allow caching to be fully disabled (run in-place with no dir proxy) for CI/CD
    * Cache should have allow a timeout parameter


decorator used to implement caching:
    * decorator args:
        * inputs:       list or glob
        * outputs       (do we need this, can we just detect this??)
                        Default to "." (everything)
        * timeout=0     Seconds before cache should be considered stale
        * name=None     If not given, default to the short name of the function.
                        (Cache "slot"), must be filesystem safe]
"""

from __future__ import annotations

from functools import wraps
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import List, Optional, Union

from ksconf.builder import QUIET, VERBOSE, BuildCacheException, BuildStep
from ksconf.builder.cache import CachedRun, FileSet


def _get_function_sourcecode_hash(f):
    import hashlib
    import inspect
    h = hashlib.new("sha256")
    code = inspect.getsource(f).encode("utf-8")
    h.update(code)
    return h.hexdigest()


class BuildManager:
    """ Supports an application building process by managing individual build steps

    .. versionadded:: v0.8.0
    """

    def __init__(self):
        self.source_path: Optional[Path] = None
        self.build_path: Optional[Path] = None
        self.dist_path: Optional[Path] = None
        self.cache_path = None
        self._cache_enabled = True
        self._taint = False
        self._folder_set = False

    def taint_cache(self):
        self._taint = True

    def disable_cache(self):
        self._cache_enabled = False

    def get_build_step(self, output=None) -> BuildStep:
        kw = {}
        if output:
            kw["output"] = output
        step = BuildStep(self.build_path, self.source_path, self.dist_path, **kw)
        return step

    def set_folders(self,
                    source_path: Path,
                    build_path: Path,
                    dist_path: Optional[Path] = None):
        self._folder_set = True
        self.source_path = Path(source_path).absolute()
        self.build_path = Path(build_path).absolute()
        if dist_path:
            self.dist_path = Path(dist_path).absolute()
        else:
            # Assume 'dist' lives along side 'build'
            self.dist_path = self.build_path.with_name("dist")
        self.cache_path = self.build_path.with_suffix(".cache")

    def is_folders_set(self):
        return self._folder_set

    def get_cache_info(self, name: str) -> CachedRun:
        path = self.cache_path / name
        cache_info = CachedRun(path)
        if self._taint:
            cache_info.taint()
        if not self._cache_enabled:
            cache_info.disable()
        return cache_info

    def cache(self, inputs: List[str], outputs: int,
              timeout: Optional[int] = None,
              name: Optional[str] = None,
              cache_invalidation: Union[dict, list, str, None] = None) -> None:
        """ function decorator for caching build steps
        Wrapped function must accept BuildStep instance as first parameters

        XXX:  Clearly document what things are good cache candidates and which are not.

        Example:

            * No extra argument to the function (at least currently)
            * Changes to inputs files are not supported
            * Deleting files aren't supported
            * Can only operate in a single directory given a limited set of inputs
            * Cannot read from the source directory, and agrees not to write to dist
              (In other words, limit all activities to build_path for deterministic behavior)
        """
        name_ = name
        cache_settings = {
            "name": name,
            "inputs": inputs,
            "outputs": outputs,
            "timeout": timeout,
            "cache_invalidation": cache_invalidation,
        }

        def decorator(f):
            # nonlocal name
            if name_ is None:
                name = f.__name__.split(".")[-1]

            f_source_hash = _get_function_sourcecode_hash(f)
            cache_settings["function_code_hash"] = f_source_hash

            @wraps(f)
            def wrapper(build_step: BuildStep) -> None:
                log = build_step.get_logger(name)
                use_cache = True
                cache = self.get_cache_info(name)

                # After the first cached build step is executed, block access to source path
                build_step.source_path = None

                if cache.is_disabled:
                    log("Caching disabled", VERBOSE)
                    # Run directly with no additional
                    return f(build_step)

                if cache.exists:
                    try:
                        cache.load()
                    except Exception as e:
                        log(f"Failed to load cache:  {e}", QUIET)
                        use_cache = False
                if cache._settings:
                    # Always use the most up-to-date value for timeout
                    cache.set_settings({"timeout": timeout})
                    try:
                        for setting in cache_settings:
                            if cache_settings[setting] != cache._settings[setting]:
                                log(f"Cache invalided due to setting '{setting}': "
                                    f"{cache_settings[setting]} was "
                                    f"{cache._settings[setting]}")
                                use_cache = False
                                break
                    except KeyError as e:
                        log(f"Cache invalided due to missing setting {e}")
                        use_cache = False
                # TODO: Check for cache tampering (confirm that existing files haven't
                #       been modified); user requestable
                current_inputs = FileSet.from_filesystem(self.source_path, inputs)
                # Determine if previous cache entry exists, and if the input files are the same
                if not cache.exists:
                    log("No cache found", VERBOSE)
                    use_cache = False
                elif not cache.inputs_identical(current_inputs):
                    # XXX: Tell user which file(s) changed?
                    log("Cache skipped due to change in inputs.  Will re-run.")
                    use_cache = False
                elif cache.is_expired:
                    log("Cache expired.  Will re-run.")
                    use_cache = False
                if use_cache:
                    # Cache HIT:  Reuse cache, by simply copying the outputs to the build folder
                    log("Cache used")
                    cached_output = cache.cached_outputs
                    cached_output.copy_all(cache.cache_dir, self.build_path)
                    log(f"Reused {len(cached_output)} output objects from cache", VERBOSE)
                    # XXX:  VERBOSE 3 should list all expanded files
                else:
                    # Cache MISS: Prepare to call the wrapped function
                    # 1. √ Make temporary location
                    # 2. √ Copy inputs to temp locations
                    # 3. √ Run wrapped functionality
                    # 4. √ Expand output into FileSet (Collect)
                    # 5. √ Store metadata (inputs/outputs/...) to JSON
                    # 6. √ Copy output to real build folder

                    # Make temporary folder for executing wrapped function
                    with TemporaryDirectory(dir=self.cache_path,
                                            prefix=f"{name}-tmp-") as temp_dir:
                        # NOTE: Make a 't' dir under the temp folder so that the cache.rename() doesn't remove the
                        #       folder managed by TemporaryDirectory, which it doesn't like.
                        temp_dir = Path(temp_dir) / "t"
                        final_cache_root = cache.root
                        cache = CachedRun(temp_dir)
                        alt_bs = build_step.alternate_path(cache.cache_dir)
                        # XXX: Copy any other settings from the original 'build_step' to our copied version
                        # Collect inputs from the source directory and copy them to the temporary directory
                        fs_inputs = FileSet.from_filesystem(self.source_path, inputs)
                        fs_inputs.copy_all(self.source_path, cache.cache_dir)
                        log(f"Copied {len(fs_inputs)} input files", VERBOSE * 2)
                        log(f"Copied input files: {', '.join(str(p) for p in fs_inputs)}", VERBOSE * 3)
                        try:
                            # Run wrapped function
                            ret = f(alt_bs)
                            if ret is not None:
                                raise NotImplementedError("A return value not supported for cached build steps")
                        except Exception as e:
                            log(f"Failed during executing.  Error: {e}", QUIET * 2)
                            # XXX: More error handling here.  Add custom exception class
                            raise
                        # TODO: Check for change to inputs.  This could lead to nondeterministic results
                        fs_inputs2 = FileSet.from_filesystem(cache.cache_dir, inputs)
                        if fs_inputs != fs_inputs2:
                            log("Inputs changed during execution", QUIET * 2)
                            raise BuildCacheException("Inputs were modified")
                        fs_outputs = FileSet.from_filesystem(cache.cache_dir, outputs)
                        # Store input/output fingerprint to the internal cache state and persist to disk
                        log(f"Capture {len(fs_outputs)} outputs")
                        cache.set_settings(cache_settings)
                        cache.set_cache_info("inputs", fs_inputs)
                        cache.set_cache_info("outputs", fs_outputs)
                        cache.dump()
                        # Copy output files to real build directory
                        fs_outputs.copy_all(cache.cache_dir, self.build_path)
                        cache.rename(final_cache_root)
                # No return (on purpose); as we don't want to track extra state

            wrapper.inputs = inputs
            wrapper.outputs = outputs
            wrapper.timeout = timeout
            wrapper.name = name
            return wrapper
        return decorator



from __future__ import absolute_import, unicode_literals

import filecmp
import os
import re
import shutil
import sys
from contextlib import contextmanager
from glob import glob
from io import open
from pathlib import Path
from random import randint
from typing import IO, Callable, Iterable, List, Tuple, Union

from ksconf.consts import SMART_CREATE, SMART_NOCHANGE, SMART_UPDATE, is_debug
from ksconf.types import PathType
from ksconf.util.compare import file_compare


def _is_binary_file(filename, peek=2048):
    # https://stackoverflow.com/a/7392391/315892
    textchars = bytearray({7, 8, 9, 10, 12, 13, 27} | set(range(0x20, 0x100)) - {0x7f})
    with open(filename, "rb") as f:
        b = f.read(peek)
        return bool(b.translate(None, textchars))


_dir_exists_cache = set()


def dir_exists(directory):
    """ Ensure that the directory exists """
    # This works as long as we never call os.chdir()
    if directory in _dir_exists_cache:
        return
    if not os.path.isdir(directory):
        os.makedirs(directory)
    _dir_exists_cache.add(directory)


def smart_copy(src, dest):
    """ Copy (overwrite) file only if the contents have changed. """
    ret = SMART_CREATE
    if os.path.isfile(dest):
        if file_compare(src, dest):
            # Files already match.  Nothing to do.
            return SMART_NOCHANGE
        else:
            ret = SMART_UPDATE
            os.unlink(dest)
    shutil.copy2(src, dest)
    return ret


def _stdin_iter(stream=None):
    if stream is None:
        stream = sys.stdin
    for line in stream:
        yield line.rstrip()


def file_fingerprint(path, compare_to=None):
    stat = os.stat(path)
    fp = (stat.st_mtime, stat.st_size)
    if compare_to:
        return fp != compare_to
    else:
        return fp


def expand_glob_list(iterable, do_sort=False):
    for item in iterable:
        if "*" in item or "?" in item:
            glob_expanded = glob(item)
            if do_sort:
                glob_expanded.sort()
            for match in glob_expanded:
                yield match
        else:
            yield item


# This is a Splunk-style (props) stanza style glob:
# where '*' is a single path component, and '...' or '**' means any depth
_glob_to_regex = [
    ("**", r".*"),
    ("*", r"[^/\\]*"),
    ("?", r"."),
    ("...", r".*"),
    (".", r"\."),
]
_glob_to_regex_find = "({})".format("|".join(re.escape(r) for r, _ in _glob_to_regex))


def splglob_to_regex(pattern, re_flags=None):
    glob_to_regex = dict(_glob_to_regex)
    regex = re.sub(_glob_to_regex_find, lambda m: glob_to_regex[m.group()], pattern)
    # If NO anchors have been explicitly given, then assume full-match mode:
    if not re.search(r'(?<![\[\\])[$^]', regex):
        regex = f"^{regex}$"
    return re.compile(regex, flags=re_flags)


def splglob_simple(pattern):
    """ Return a splglob that either matches a full path or match a simple file """
    if "/" not in pattern:
        # Assume we've been given a simple file name:   app.conf, *.tgz
        pattern = f"^.../{pattern}$"
    else:
        pattern = f"^{pattern}$"
    return pattern


def relwalk(top: PathType,
            topdown=True, onerror=None, followlinks=False
            ) -> Iterable[Tuple[str, List[str], List[str]]]:
    """ Relative path walker
    Like os.walk() except that it doesn't include the "top" prefix in the resulting 'dirpath'.
    """
    top = os.fspath(top)
    if not top.endswith(os.path.sep):
        top += os.path.sep
    prefix = len(top)
    for (dirpath, dirnames, filenames) in os.walk(top, topdown, onerror, followlinks):
        dirpath = dirpath[prefix:]
        yield (dirpath, dirnames, filenames)


def file_hash(path: PathType, algorithm="sha256") -> str:
    import hashlib
    h = hashlib.new(algorithm)
    with open(path, "rb") as fp:
        buf = True
        while buf:
            buf = fp.read(4096)
            h.update(buf)
    return h.hexdigest()


def _samefile(file1: PathType, file2: PathType) -> bool:
    if hasattr(os.path, "samefile"):
        # Nix
        return os.path.samefile(file1, file2)
    else:
        # Windows
        file1 = os.path.normpath(os.path.normcase(file1))
        file2 = os.path.normpath(os.path.normcase(file2))
        return file1 == file2


def secure_delete(path: Path, passes=3):
    """
    A simple file shred technique.  If there's demand, this could be expanded.
    But for now, 'secure' means just slightly more secure that unlink().

    Adapted from from Ansible's _shred_file_custom()
    """
    path = Path(path)
    file_len = path.stat().st_size
    if file_len == 0:
        # avoid work when empty
        return
    max_chunk_len = min(1024 * 1024 * 2, file_len)
    with open(path, "wb") as fh:
        for _ in range(passes):
            fh.seek(0, 0)
            # get a random chunk of data, each pass with other length
            chunk_len = randint(max_chunk_len // 2, max_chunk_len)
            data = os.urandom(chunk_len)

            for _ in range(0, file_len // chunk_len):
                fh.write(data)
            fh.write(data[:file_len % chunk_len])
            os.fsync(fh)
    path.unlink()


@contextmanager
def atomic_writer(dest: Path,
                  temp_name: Union[Path, str, Callable[[Path], Path], None]) -> Path:
    """
    Context manager to atomically update a destination.  When entering the context, a temporary file
    name is returned.  When the context is successfully exited, the temporary file is renamed into
    place.  Either way, the temporary file is removed.

    The name of the temporary file can be controlled via ``temp_name``.  If a ``str`` is provided,
    it will be used as a suffix.  If a Path is provided, that will be used as the literal temporary
    file name.  If a callable is given, the ``dest`` path will be passed into the callable to
    determine the temporary file.  Alternatively, the entire _atomic_ nature of this function can be
    disabled by passing temp_name=None.
    """
    dest = Path(dest)

    if temp_name is None:
        # Opt-out of temp file handling.  Just write directly to the real destination.
        yield dest
        return

    if isinstance(temp_name, str):
        # Suffix mode:  Concatenate 'temp_name' to existing name
        if not temp_name.startswith("."):
            temp_name = "." + temp_name
        temp_dest: Path = dest.with_name(dest.name + temp_name)
    elif isinstance(temp_name, Path):

        # The actual temp file name was created by the caller
        temp_dest = temp_name
    elif callable(temp_name):
        # Temp filename to be generated via function call
        temp_dest = Path(temp_name(dest))
    else:
        raise TypeError("Unsupported type for 'temp_name'.  "
                        f"Expected str, Path, or callable but received {type(temp_name).__name__}")

    if dest == temp_dest:
        # If you want to write directly to the output file, use temp_name=None
        # Having dest==temp_dest will result in all files being removed.
        raise ValueError(f"Invalid options.  Both dest and temp_dest are the same:  {dest}")

    # Remove any existing temporary files.
    if temp_dest.is_file():
        temp_dest.unlink()

    try:
        yield temp_dest
        assert temp_dest.is_file(), f"No file written to the temporary path {temp_dest}, " \
                                    f"therefore we cannot rename it to {dest}"
        if dest.is_file():
            dest.unlink()
        temp_dest.replace(dest)
    finally:
        if temp_dest.is_file():
            temp_dest.unlink()


@contextmanager
def atomic_open(name: Path,
                temp_name: Union[Path, str, Callable[[Path], Path], None],
                mode="w",
                **open_kwargs) -> IO:
    """
    Context manager to atomically write to a file stream.  Like the open() context manager, a file
    handle returned when the context is entered.  Upon successful completion, the temporary file is
    renamed into place; thus providing an atomic update operation.

    See :func:`atomic_writer` for behaviors regarding the ``temp_name`` parameter option.

    This function can be used nearly any place that ``with open(myfile, mode="w") as stream``
    """
    with atomic_writer(name, temp_name) as tmp_filename:
        with open(tmp_filename, mode=mode, **open_kwargs) as stream:
            yield stream


class ReluctantWriter:
    """
    Context manager to intelligently handle updates to an existing file.  New content is written
    to a temp file, and then compared to the current file's content.  The file file will be
    overwritten only if the contents changed.
    """

    def __init__(self, path, *args, **kwargs):
        self.path = path
        self._arg = (args, kwargs)
        self._fp = None
        self._tmpfile = path + ".tmp"
        self.change_needed = None
        self.result = None

    def __enter__(self):
        args, kwargs = self._arg
        self._fp = open(self._tmpfile, *args, **kwargs)
        return self._fp

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Don't do anything, other than try to close/delete the file, if an error occurred.
        try:
            self._fp.close()
        except Exception:
            raise
        if exc_type:
            if is_debug():
                # LOG that temp file is being kept
                pass
            else:
                os.unlink(self._tmpfile)
            return
        if not os.path.isfile(self.path):
            os.rename(self._tmpfile, self.path)
            self.change_needed = True
            self.result = "created"
        elif filecmp.cmp(self._tmpfile, self.path):
            os.unlink(self._tmpfile)
            self.change_needed = False
            self.result = "unchanged"
        else:
            os.unlink(self.path)
            os.rename(self._tmpfile, self.path)
            self.change_needed = True
            self.result = "updated"



from __future__ import annotations

import re
from collections import defaultdict
from dataclasses import dataclass, field
from fnmatch import fnmatch
from os import PathLike, stat_result
from pathlib import Path, PurePath
from tempfile import NamedTemporaryFile
from typing import Callable, Iterator, Match, Optional

from ksconf.compat import Dict, List, Set, Tuple
from ksconf.hook import plugin_manager
from ksconf.util.file import relwalk, secure_delete

"""

LayerRootBase has one or more 'Layer', each layer has one or more 'File's.


LayerRoot methods:

    - list_files():   Return superset of all file names returned by all layers (order undefined)
    - walk():         Return list_files() like content in a os.walk() (or relwalker) like way --
                      must consider directory order, useful for copying tree, for example.
                      Assumption for now:  Don't return layer per file, just what files exist.
                      Ask about layers per file later.
    - list_layers():  Iterate over layer objects (metadata retrievable, on demand)
    - get_file():     Return files (in ranked layer order)


Other possible methods:
    list_dirs():      Return list of known directories?   Not sure how we want this part to work.
                      Perhaps walk() is good enough?


dotD style layers:

    LayerRootDotD   has one or more 'LayerMount', each LayerMount has one or more Layer,
                    which has one or more 'File's.



Remember:  This must work for NON-layered directories too, hopefully with minimal overhead.

This must work with an explicitly given list of layers

"""


def _path_join(*parts):
    """ A slightly smarter / more flexible path appender.
    Drop any None
    """
    return Path(*filter(None, parts))


# Exceptions

class LayerException(Exception):
    pass


class LayerUsageException(LayerException):
    pass


@dataclass
class LayerContext:
    follow_symlink: bool = False
    block_files: Match = re.compile(r"\.(bak|swp)$")
    block_dirs: set = field(default_factory=lambda: {".git"})
    template_variables: dict = field(default_factory=dict)


@dataclass
class _FileFactoryHandler:
    name: str
    handler: LayerFile
    priority: int = 0
    enabled: bool = False


class FileFactory:
    _registered_handlers: Dict[str, _FileFactoryHandler] = {}
    _enabled_handlers: List[Callable] = []

    def __init__(self):
        # Make this class a singleton?
        self._context_state = None

    def _recalculate(self):
        self._enabled_handlers = [
            h.handler for h in sorted(self._registered_handlers.values(),
                                      key=lambda h: (-h.priority, h.name)) if h.enabled]

    def enable(self, name, _enabled=True):
        handler = self._registered_handlers[name]
        handler.enabled = _enabled
        self._recalculate()

    def disable(self, name):
        self.enable(name, False)

    def list_available_handlers(self) -> List[str]:
        return [h.name for h in self._registered_handlers.values() if not h.enabled]

    def __call__(self, layer, path: PurePath, *args, **kwargs) -> LayerFile:
        """
        Factory thats finds the appropriate LayerFile class and returns a new instance.
        """
        for cls in self._enabled_handlers:
            if cls.match(path):
                return cls(layer, path, *args, **kwargs)

    def register_handler(self, name: str, **kwargs):
        def wrapper(handler_class: LayerFile):
            handler = _FileFactoryHandler(name, handler_class, **kwargs)
            self._registered_handlers[handler.name] = handler
            self._recalculate()
            return handler_class
        return wrapper

    def __enter__(self) -> FileFactory:
        # The primary use case is for clean unit testing
        assert not self._context_state, "Nested contexts are not supported"
        from copy import deepcopy
        self._context_state = (deepcopy(self._registered_handlers),
                               deepcopy(self._enabled_handlers))
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._registered_handlers, self._enabled_handlers = self._context_state
        self._context_state = None


# Single shared instance
layer_file_factory = FileFactory()
register_file_handler = layer_file_factory.register_handler


@register_file_handler("default", priority=-100, enabled=True)
class LayerFile(PathLike):
    '''
    Abstraction of a file within a Layer

    Path definitions

    ..

    ``logical_path``
        Conceptual file path.  This is the final path after all layers are resolved.
        Think of this as the 'destination' file.

    ``physical_path``
        Actual file path.  The location of the physical file found within a source layer.
        Most of the time this is the 'source' file, however this doesn't take into considerations layer combining or
        template expansion requirements.  (In the case of a template, this would be the template file)

    ``resource_path``
        Content location.  Often this the ``physical_path``, but in the case of abstracted layers
        (like templates, or archived layers), this would be the location of a temporary resource that contains
        the expanded/rendered content.
    '''
    __slots__ = ["layer", "relative_path", "_stat"]

    def __init__(self,
                 layer: LayerRootBase.Layer,
                 relative_path: PurePath,
                 stat: Optional[stat_result] = None):
        self.layer = layer
        self.relative_path = relative_path
        self._stat = stat

    def __fspath__(self) -> str:
        return self.resource_path

    @staticmethod
    def match(path: PurePath):
        return True

    @property
    def physical_path(self) -> Path:
        return _path_join(self.layer.root, self.layer.physical_path, self.relative_path)

    @property
    def logical_path(self) -> Path:
        return _path_join(self.layer.logical_path, self.relative_path)

    # For "normal" files, the resource_path is the physical_path (not true for rendered files)
    resource_path = physical_path

    @property
    def stat(self) -> stat_result:
        if self._stat is None:
            self._stat = self.physical_path.stat()
        return self._stat

    @property
    def size(self):
        return self.stat.st_size

    @property
    def mtime(self):
        return self.stat.st_mtime


class LayerRenderedFile(LayerFile):
    """
    Abstract LayerFile for rendered scenarios, such as template scenarios.
    A subclass really only needs to implement ``match()`` ``render()``
    """
    __slots__ = ["_rendered_resource"]

    use_secure_delete = False

    def __init__(self, *args, **kwargs):
        super(LayerRenderedFile, self).__init__(*args, **kwargs)
        self._rendered_resource = None

    def __del__(self):
        if getattr(self, "_rendered_resource", None) and self._rendered_resource.is_file():
            if self.use_secure_delete:
                # Use (slightly-more) secure deletion.
                # Note that in a packaging operation, for example, there are many temporary files
                # that could contain sensitive data.  This is very imperfect.
                secure_delete(self._rendered_resource)
            else:
                self._rendered_resource.unlink()

    def render(self, template_path: Path) -> str:
        raise NotImplementedError

    @staticmethod
    def transform_name(path: PurePath):
        # Remove trailing suffix
        return path.with_name(path.stem)

    @property
    def logical_path(self) -> Path:
        return _path_join(self.layer.logical_path,
                          self.transform_name(self.relative_path))

    @property
    def physical_path(self) -> Path:
        return _path_join(self.layer.root, self.layer.physical_path, self.relative_path)

    @property
    def resource_path(self) -> Path:
        if not self._rendered_resource:
            # Temporary file will be removed in instance destructor.  Multiple opens expected.
            tf = NamedTemporaryFile(delete=False)
            self._rendered_resource = Path(tf.name)
            content = self.render(self.physical_path)
            self._rendered_resource.write_text(content)
        return self._rendered_resource


@register_file_handler("jinja", priority=50, enabled=False)
class LayerFile_Jinja2(LayerRenderedFile):
    @staticmethod
    def match(path: PurePath):
        return path.suffix == ".j2"

    @staticmethod
    def transform_name(path: PurePath):
        return path.with_name(path.name[:-3])

    @property
    def jinja2_env(self):
        # Use context object to 'cache' the jinja2 environment
        if not hasattr(self.layer.context, "jinja2_environment"):
            self.layer.context.jinja2_environment = self._build_jinja2_env()
        return self.layer.context.jinja2_environment

    def _build_jinja2_env(self):
        from jinja2 import Environment, FileSystemLoader, StrictUndefined
        environment = Environment(
            undefined=StrictUndefined,
            loader=FileSystemLoader(self.layer.root),
            auto_reload=False)

        # Call plugin for jinja environment tweaking
        plugin_manager.hook.modify_jinja_env(env=environment)

        environment.globals.update(self.layer.context.template_variables)
        return environment

    def render(self, template_path: Path) -> str:
        rel_template_path = template_path.relative_to(self.layer.root)
        template = self.jinja2_env.get_template("/".join(rel_template_path.parts))
        value = template.render()
        return value


class LayerFilter:
    _valid_actions = ("include", "exclude")

    def __init__(self):
        self._rules = []

    def add_rule(self, action, pattern):
        # If no filter rules have been setup yet, be sure to set the default
        if action not in self._valid_actions:
            raise ValueError(f"Unknown action of {action}.  "
                             f"Valid actions include: {self._valid_actions}")
        if not self._rules:
            if action == "include":
                first_filter = ("exclude", "*")
            elif "exclude":
                first_filter = ("include", "*")
            self._rules.append(first_filter)
        self._rules.append((action, pattern))

    def evaluate(self, layer: LayerRootBase.Layer) -> bool:
        response = True
        layer_name = layer.name
        for rule_action, rule_pattern in self._rules:
            if fnmatch(layer_name, rule_pattern):
                response = rule_action == "include"
        return response

    __call__ = evaluate


R_walk = Iterator[Tuple[Path, List[str], List[str]]]


class LayerRootBase:
    """ All 'path's here are relative to the ROOT. """

    class Layer:
        """ Basic layer Container:   Connects logical and physical paths. """
        __slots__ = ["name", "root", "logical_path", "physical_path", "context",
                     "_file_factory", "_cache_files"]

        def __init__(self, name: str,
                     root: Path,
                     physical: PurePath,
                     logical: PurePath,
                     context: LayerContext,
                     file_factory: Callable):
            self.name = name
            self.root = root
            self.physical_path = physical
            self.logical_path = logical
            self.context = context
            self._file_factory = file_factory
            self._cache_files: List[LayerFile] = []

        def walk(self) -> R_walk:
            # In the simple case, this is good enough.   Some subclasses will need to override
            for (root, dirs, files) in relwalk(_path_join(self.root, self.physical_path),
                                               followlinks=self.context.follow_symlink):
                root = Path(root)
                files = [f for f in files if not self.context.block_files.search(f)]
                for d in list(dirs):
                    if d in self.context.block_dirs:
                        dirs.remove(d)
                yield (root, dirs, files)

        def iter_files(self) -> Iterator[LayerFile]:
            for (top, _, files) in self.walk():
                for file in files:
                    yield self._file_factory(self, top / file)

        def list_files(self) -> List[LayerFile]:
            if not self._cache_files:
                self._cache_files = list(self.iter_files())
            return self._cache_files

        def get_file(self, path: Path) -> LayerFile:
            """ Return file object (by logical path), if it exists in this layer. """
            # TODO:  Optimize by caching.  Use a dict with a logical_path as the key
            for file in self.list_files():
                if file.logical_path == path:
                    if file.physical_path.is_file():
                        return file
                    else:
                        return None

    # LayerRootBase
    def __init__(self, context: Optional[LayerContext] = None):
        self._layers: List[LayerRootBase.Layer] = []
        self.context = context or LayerContext()

    def apply_filter(self, layer_filter: LayerFilter) -> bool:
        """
        Apply a destructive filter to all layers.  ``layer_filter(layer)`` will be called one for each
        layer, if the filter returns True than the layer is kept.  Root layers are always kept.

        Returns True if layers were removed
        """
        layers = [l for l in self._layers if layer_filter(l)]
        result = self._layers != layers
        self._layers = layers
        return result

    def order_layers(self):
        raise NotImplementedError

    def add_layer(self, layer: Layer, do_sort=True):
        self._layers.append(layer)
        if do_sort:
            self.order_layers()

    def list_layers(self) -> List[Layer]:
        return self._layers

    def get_layers_by_name(self, name: str) -> Iterator[LayerRootBase.Layer]:
        for layer in self.list_layers():
            if layer.name == name:
                yield layer

    def list_layer_names(self) -> List[str]:
        return [l.name for l in self.list_layers()]

    def iter_all_files(self) -> Iterator[LayerFile]:
        """ Iterator over all physical files. """
        for layer in self._layers:
            yield from layer.iter_files()

    def list_physical_files(self) -> List[LayerFile]:
        files = set()
        for file_ in self.iter_all_files():
            files.add(file_.physical_path)
        return list(files)

    def list_logical_files(self) -> List[LayerFile]:
        """ Return a list of logical paths. """
        files = set()
        for file_ in self.iter_all_files():
            files.add(file_.logical_path)
        return list(files)

    def get_file(self, path) -> Iterator[LayerFile]:
        """ return all layers associated with the given relative path. """
        for layer in self._layers:
            file_ = layer.get_file(path)
            if file_:
                yield file_

    # Legacy names
    list_files = list_logical_files


class DirectLayerRoot(LayerRootBase):
    """
    A very simple direct LayerRoot implementation that relies on all layer paths to be explicitly
    given without any automatic detection mechanisms.  You can think of this as the legacy
    implementation.
    """

    def add_layer(self, path: Path):
        Layer = self.Layer
        # Layer name should be considered arbitrary and unimportant here
        layer_name = path.name
        if not path.is_dir():
            raise LayerUsageException("Layers must be directories.  "
                                      f"Given path '{path}' is not a directory.")
        layer = Layer(layer_name, path, None, None, context=self.context,
                      file_factory=layer_file_factory)
        super(DirectLayerRoot, self).add_layer(layer)

    def order_layers(self):
        # No op.  Irrelevant as layers are given (CLI) in the order they should be applied.
        pass


"""


    Layers must be discovered depth-first,
    Files must be walked breath-first?  Is that right?

    MyApp/                          <- Root + MountPointTransparent (contains 1 layer)
        bin/
        default.d/                  <- MountPoint Disconnect from parent layer
            10-upstream             <- Layer
            20-kintyre-core         <- Layer
            99-override             <- Layer
        static/
        lookup.d/



Q:  What about nested layers, should that be supported?  (Yes, this is an horrific example)
A:  Let's support if it just works naturally, otherwise, let's not go to extra lengths to support
A:  Multiple LayerRoots SHOULD be supported.

    MyApp/                          <- LayerRoot & Layer (Anonymous -- lowest ranking)
        lib.d/
            10-upstream
                botocore/
                requests/
            20-kintyre              <-- LayerMount
                botocore.d/
                    10-python3      <-- Layer (nested)
                    20-python2      <-- Layer (nested)
                requests/
        default.d/                  <- Disconnect from parent layer
            10-upstream             <- Layer
            20-kintyre-core         <- Layer
            99-override             <- Layer
        static/

"""


# Q:  How do we mark "mount-points" in the directory structure to keep multiple layers
#     from claiming the same files?????
class DotDLayerRoot(LayerRootBase):

    class Layer(LayerRootBase.Layer):
        __slots__ = ["prune_points"]

        def __init__(self, name: str,
                     root: Path,
                     physical: PurePath,
                     logical: PurePath,
                     context: LayerContext,
                     file_factory: Callable,
                     prune_points: Set[Path] = None):
            super(DotDLayerRoot.Layer, self).__init__(name, root, physical, logical, context=context,
                                                      file_factory=file_factory)
            self.prune_points: Set[Path] = set(prune_points) if prune_points else set()

        def walk(self) -> R_walk:
            for (root, dirs, files) in super(DotDLayerRoot.Layer, self).walk():
                if root in self.prune_points:
                    # Cleanup files/dirs to keep walk() from descending deeper
                    del dirs[:]
                else:
                    yield (root, dirs, files)

    '''
    class MountBase:
        def __init__(self, path):
            self.path = path

    class MountTransparent(MountBase):
        """ Pass through files as-is, no manipulation. """
        pass

    class MountDotD(MountBase):
        def __init__(self, path):
            super(DotDLayerRoot.MountDotD, self).__init__(path)
    '''

    mount_regex = re.compile(r"(?P<realname>[\w_.-]+)\.d$")
    layer_regex = re.compile(r"(?P<layer>\d\d-[\w_.-]+)")

    def __init__(self, context=None):
        super(DotDLayerRoot, self).__init__(context)
        # self.root = None
        self._root_layer: LayerRootBase.Layer = None
        self._mount_points: Dict[Path, List[str]] = defaultdict(list)

    def apply_filter(self, layer_filter: LayerFilter):
        # Apply filter function, but also be sure to keep the root layer
        def fltr(l):
            return l is self._root_layer or layer_filter(l)
        return super(DotDLayerRoot, self).apply_filter(fltr)

    def set_root(self, root: Path, follow_symlinks=None):
        """ Set a root path, and auto discover all '.d' directories.

        Note:  We currently only support ``.d/<layer>`` directories, a file like
        ``default.d/10-props.conf`` won't be handled here.
        A valid name would be ``default.d/10-name/props.conf``.
        """
        Layer = self.Layer
        root = Path(root)
        if follow_symlinks is None:
            follow_symlinks = self.context.follow_symlink

        for (top, dirs, files) in relwalk(root, topdown=False, followlinks=follow_symlinks):
            del files
            top = Path(top)
            mount_mo = self.mount_regex.match(top.name)
            if mount_mo:
                for dir_ in dirs:
                    dir_mo = self.layer_regex.match(dir_)
                    if dir_mo:
                        # XXX: Nested layers breakage, must substitute multiple ".d" folders in `top`
                        layer = Layer(dir_mo.group("layer"),
                                      root,
                                      physical=top / dir_,
                                      logical=top.parent / mount_mo.group("realname"),
                                      context=self.context,
                                      file_factory=layer_file_factory)
                        self.add_layer(layer)
                        self._mount_points[top].append(dir_)
                    else:
                        # XXX: Give the user the option of logging the near-matches (could indicate a
                        # problem in the config, or could be some other legit directory structure)
                        '''
                        print(f"LAYER NEAR MISS:  {top} looks like a mount point, but {dir_} doesn't "
                              "follow the expected convention")
                        '''
                        pass
            elif top.name.endswith(".d"):
                '''
                print(f"MOUNT NEAR MISS:  {top}")
                '''
                pass

        # XXX: Adding <root> should be skipped if (and only if) root itself if a '.d' folder
        # Very last operation, add the top directory as the final layer (lowest rank)
        prune_points = [mount / layer
                        for mount, layers in self._mount_points.items()
                        for layer in layers]
        layer = Layer("<root>", root, None, None, context=self.context,
                      file_factory=layer_file_factory,
                      prune_points=prune_points)
        self.add_layer(layer, do_sort=False)
        self._root_layer = layer

    def list_layers(self) -> List[Layer]:
        # Return all but the root layer.
        # Avoiding self._layers[:-1] because there could be cases where root isn't included.
        return [l for l in self._layers if l is not self._root_layer]

    def order_layers(self):
        # Sort based on layer name (or other sorting priority:  00-<name> to 99-<name>
        self._layers.sort(key=lambda l: l.name)



import os
import sys

from pluggy import HookimplMarker, PluginManager

from ksconf.consts import EXIT_CODE_BROKEN_PLUGIN, PLUGGY_DISABLE_HOOK, PLUGGY_HOOK, is_debug
from ksconf.hookspec import KsconfHookSpecs

# decorator to mark ksconf hooks
ksconf_hook = HookimplMarker(PLUGGY_HOOK)


class BadPluginWarning(UserWarning):
    """ Issue with one or more plugins """
    pass


class _plugin_manager:
    """
    Lazy loading PluginManager proxy that allows simple module-level access to
    'plugin_manager.hook.x' with typing.  This follows the singleton pattern.
    Everything is loaded upon first use.
    """

    def __init__(self):
        self.__plugin_manager = None
        self.__output = sys.stderr
        self.__revert_funct = None

    def log(self, message):
        self.__output.write(f"ksconf-plugin: {message}\n")

    def _startup(self):
        # This runs on demand, exactly once.
        self.__plugin_manager = pm = PluginManager(PLUGGY_HOOK)

        if PLUGGY_DISABLE_HOOK in os.environ:
            disabled_hooks = os.environ[PLUGGY_DISABLE_HOOK].split()
            for hook in disabled_hooks:
                self.log(f"Disabling {hook} by request!\n")
                pm.set_blocked(hook)

        pm.add_hookspecs(KsconfHookSpecs)

        # Disable *ALL* 'ksconf_plugin' hooks
        disabled_things = os.environ.get("KSCONF_DISABLE_PLUGINS", "").split()
        if "ksconf_plugin" in disabled_things:
            if is_debug:
                self.log("All plugins are disabled")
            return

        # XXX: Find a way to report *which* plugin failed.  For now, use traceback
        try:
            pm.load_setuptools_entrypoints("ksconf_plugin")
        except ModuleNotFoundError as e:
            self.log("Unable to load one or more ksconf plugins.  "
                     f"Please correct or uninstall the offending plugin.\n  {e}\n"
                     "Run with KSCONF_DEBUG=1 to see a traceback")
            if is_debug():
                raise e
            else:
                sys.exit(EXIT_CODE_BROKEN_PLUGIN)

        # Maybe this should be it's own setting someday?  (disconnected from KSCONF_DEBUG)
        if is_debug():
            self.enable_monitoring()

    def enable_monitoring(self):
        def before(hook_name, hook_impls, kwargs):
            if hook_impls:
                self.log(f"call start {hook_name}, impl={hook_impls} kwargs={kwargs}")

        def after(outcome, hook_name, hook_impls, kwargs):
            if hook_impls:
                self.log(f"call-end   {hook_name}, impl={hook_impls} kwargs={kwargs} "
                         f"return={outcome}\n")
        self.__revert_funct = self._plugin_manager.add_hookcall_monitoring(before, after)

    def disable_monitoring(self):
        if callable(self.__revert_funct):
            self.__revert_funct()
            return True
        return False

    @property
    def _plugin_manager(self) -> PluginManager:
        if self.__plugin_manager is None:
            self._startup()
        return self.__plugin_manager

    @property
    def hook(self) -> KsconfHookSpecs:
        return self._plugin_manager.hook

    def __getattr__(self, name: str):
        # Proxy all other content to PluginManager
        return getattr(self._plugin_manager, name)


# Shared singleton
plugin_manager = _plugin_manager()


def get_plugin_manager() -> _plugin_manager:
    """
    Return the shared pluggy PluginManager (singleton) instance.

    This is for backwards compatibility.  This was only added in v0.11.6; and replaced immediately after.
    """
    from warnings import warn
    warn("Please use 'plugin_manager' directly and not 'get_plugin_manager()'. "
         "This function will be removed in v0.13 or sooner.", DeprecationWarning)
    return plugin_manager


if __name__ == '__main__':
    print(plugin_manager.list_name_plugin())



""" SUBCOMMAND:  ``ksconf package -f <SPL> <DIR>``

Usage example:

.. code-block:: sh

   ksconf package -f myapp.tgz MyApp/


Build system example:

.. code-block:: sh

    ksconf package -f release/myapp-{{version}}.tgz \\
            --block-local \\
            --set-version={{git_tag}} \\
            --set-build=${TRAVIS_BUILD_NUMBER:-0}


"""

import argparse
import json
import os

from ksconf.command import KsconfCmd, add_file_handler, dedent
from ksconf.consts import EXIT_CODE_BAD_ARGS, EXIT_CODE_SUCCESS
from ksconf.layer import layer_file_factory
from ksconf.package import AppPackager


class PackageCmd(KsconfCmd):
    help = "Create a Splunk app .spl file from a source directory"
    description = dedent("""\
    Create a Splunk app or add on tarball (``.spl``) file from an app directory.

    ``ksconf package`` can do useful things like, exclude unwanted files, combine layers, set the
    application version and build number, drop or promote the ``local`` directory into ``default``.

    Note that some arguments, like the ``FILE`` support special values that can be automatically
    evaluated at runtime.  For example the placeholders ``{{version}}`` or ``{{git_tag}}`` can be
    expanded into the output tarball filename.

    If both layering and templating are in use at the same time, be aware that templates are
    rendered prior to layering operations.  This allows, for example, one layer to include a simple
    ``indexes.conf`` file and another layer to include an ``indexes.conf.j2`` template.
    """)
    # format = "manual"
    maturity = "beta"

    default_blocklist = [
        ".git*",
        "*.py[co]",
        "__pycache__",
        ".DS_Store"
    ]

    def register_args(self, parser: argparse.ArgumentParser):
        def wb_type(action):
            def f(pattern):
                return action, pattern
            return f

        parser.add_argument("source", metavar="SOURCE",
                            help="Source directory for the Splunk app.")
        parser.add_argument("-f", "--file", metavar="SPL",
                            help="Name of splunk app file (tarball) to create.  "
                            "Placeholder variables in ``{{var}}`` syntax can be used here.")

        parser.add_argument("--app-name",
                            help="Specify the top-level app folder name.  "
                                 "If this is not given, the app folder name is automatically "
                                 "extracted from the basename of SOURCE.  "
                                 "Placeholder variables, such as ``{{app_id}}`` can be used here.")
        parser.add_argument("--blocklist", "-b",
                            action="append",
                            default=self.default_blocklist,
                            help="Pattern for files/directories to exclude.  "
                                 "Can be given multiple times.  You can load multiple exclusions "
                                 "from disk by using ``file://path`` which can be used "
                                 "with ``.gitignore`` for example.  (Default includes: {})"
                            .format(", ".join("``{}``".format(i) for i in self.default_blocklist)))

        # XXX: This should be smarter; stacked like we support for layers -- where order matters.
        parser.add_argument("--allowlist", "-a",
                            action="append", default=[],
                            help="Remove a pattern that was previously added to the blocklist.")

        player = parser.add_argument_group(
            "Layer filtering",
            "If the app being packaged includes multiple layers, these arguments can be used to "
            "control which ones should be included in the final app file.  If no layer options "
            "are specified, then all layers will be included.")

        player.add_argument("--layer-method",
                            choices=["dir.d", "disable"],
                            default="dir.d",
                            help="Set the layer type used by SOURCE.  "
                                 "Additional description provided in in the ``combine`` command.")
        player.add_argument("-I", "--include", action="append", default=[], dest="layer_filter",
                            type=wb_type("include"), metavar="PATTERN",
                            help="Name or pattern of layers to include.")
        player.add_argument("-E", "--exclude", action="append", default=[], dest="layer_filter",
                            type=wb_type("exclude"), metavar="PATTERN",
                            help="Name or pattern of layers to exclude from the target.")

        add_file_handler(parser)
        parser.add_argument("--template-vars",
                            default=None, action="store",
                            help="Set template variables as key=value or YAML/JSON, "
                            "if filename prepend with @")

        parser.add_argument("--follow-symlink", "-l", action="store_true", default=False,
                            help="Follow symbolic links pointing to directories.  "
                                 "Symlinks to files are always followed.")

        # Set app version or extra app version.
        parser.add_argument("--set-version", metavar="VERSION",
                            help="Set application version.  By default the application version "
                                 "is read from default/app.conf.  Placeholder variables such as "
                                 "``{{git_tag}}`` can be used here.")
        parser.add_argument("--set-build", metavar="BUILD",
                            help="Set application build number.")

        plocal = parser.add_mutually_exclusive_group()
        parser.set_defaults(local="merge")
        plocal.add_argument("--allow-local",
                            dest="local", action="store_const", const="preserve",
                            help="Allow the ``local`` folder to be kept as-is  "
                                 "WARNING: "
                                 "This goes against Splunk packaging practices, and will cause "
                                 "AppInspect to fail.  "
                                 "However, this option can be useful for private package transfers "
                                 "between servers, app backups, or other admin-like tasks.")
        plocal.add_argument("--block-local",
                            dest="local", action="store_const", const="block",
                            help="Block the ``local`` folder and ``local.meta`` from the package.")
        plocal.add_argument("--merge-local",
                            dest="local", action="store_const", const="merge",
                            help="Merge any files in ``local`` into the ``default`` folder during "
                                 "packaging.  This is the default behavior.")

        pbuild = parser.add_argument_group("Advanced Build Options",
                                           "The following options are for more advanced app "
                                           "building workflows.")
        pbuild.add_argument("--release-file",
                            help="Write the path of the newly generated archive file (SPL) after "
                                 "the archive is written.  "
                                 "This is useful in build scripts when the SPL contains variables "
                                 "so the final name may not be known ahead of time.")

    @staticmethod
    def load_blocklist(path):
        with open(path) as stream:
            for line in stream:
                line = line.rstrip()
                if line and not line.startswith("#"):
                    yield line

    def pre_run(self, args):
        if "local" in args.blocklist:
            self.stderr.write("Blocking 'local' is not supported.   "
                              "Most likely you want '--block-local' instead.\n")
            return EXIT_CODE_BAD_ARGS

        blocklist = []
        for pattern in args.blocklist:
            if pattern.startswith("file://"):
                blocklist_file = pattern[7:]
                expanded = list(self.load_blocklist(blocklist_file))
                self.stderr.write(f"Extended blocklist from {blocklist_file} "
                                  f"with {len(expanded)} entries\n")
                blocklist.extend(expanded)
            else:
                blocklist.append(pattern)
        args.blocklist = [pattern for pattern in blocklist if pattern not in args.allowlist]

    def run(self, args):
        ''' Create a Splunk app/add-on .spl file from a directory '''

        # Just call combine (writing to a temporary directory) and the tar it up.  At some point
        # we should do it all in memory, but for now this good enough.  For more sophisticated
        # builds use a temp build directory.  This is what ksconf's BuildManager does.

        app_name = args.app_name
        app_name_source = "set via commandline"
        if not app_name:
            app_name = os.path.basename(args.source)
            app_name_source = "taken from source directory"
            '''
            if os.path.basename(args.source) == ".":
                app_name = os.path.basename(os.getcwd())
                app_name_source = "extracted from working directory"
            else:
                app_name = os.path.basename(args.source)
                app_name_source = "extracted from source directory"
        '''
        self.stdout.write(f"Packaging {app_name}   (App name {app_name_source})\n")

        for handler in args.enable_handler:
            layer_file_factory.enable(handler)

        template_vars = None
        if args.template_vars:
            template_vars = self.parse_extra_vars(args.template_vars, "template-vars")
            self.stdout.write(f"Using variables: \n{json.dumps(template_vars, indent=2)}\n")

        packager = AppPackager(args.source, app_name, output=self.stderr,
                               template_variables=template_vars)

        with packager:
            packager.combine(args.source, args.layer_filter,
                             layer_method=args.layer_method,
                             allow_symlink=args.follow_symlink)
            # Handle local files
            if args.local == "merge":
                packager.merge_local()
            elif args.local == "block":
                packager.block_local()
            elif args.local == "preserve":
                pass
            else:   # pragma: no cover
                raise ValueError(f"Unknown value for 'local': {args.local}")

            if args.blocklist:
                self.stderr.write(f"Applying blocklist:  {args.blocklist!r}\n")
                packager.blocklist(args.blocklist)

            if args.set_build or args.set_version:
                packager.update_app_conf(
                    version=args.set_version,
                    build=args.set_build)

            packager.check()
            # os.system(f"ls -lR {packager.app_dir}")

            dest = args.file or "{}-{{{{version}}}}.tgz".format(packager.app_name.lower().replace("-", "_"))
            archive_path = packager.make_archive(dest)
            self.stderr.write("Archive created:  file={} size={:.2f}Kb\n".format(
                os.path.basename(archive_path), os.stat(archive_path).st_size / 1024.0))

            if args.release_file:
                # Should this be expanded to be an absolute path?
                with open(args.release_file, "w") as f:
                    f.write(archive_path)

        return EXIT_CODE_SUCCESS



""" Placeholder for any internal plugins

NOTE:  This is NOT yet enabled via entrypoints
"""

from ksconf.hook import ksconf_hook

'''
@ksconf_hook
def modify_jinja_env(env):
    pass
'''

# Keep flake8 happy
del ksconf_hook



from __future__ import absolute_import, unicode_literals

from typing import Optional


def build_rest_namespace(base: str,
                         owner: Optional[str] = None,
                         app: Optional[str] = None) -> str:
    if owner and app:
        p = (base, "servicesNS", owner, app)
    elif app:
        p = (base, "servicesNS", "nobody", app)
    elif owner:
        raise ValueError("Can't specify user without app!")
    else:
        p = (base, "services")
    return "/".join(p)


def build_rest_url(base: str,
                   service: str,
                   owner: Optional[str] = None,
                   app: Optional[str] = None) -> str:
    prefix = build_rest_namespace(base, owner, app)
    return f"{prefix}/{service}"



""" SUBCOMMAND:  ``ksconf promote <SOURCE> <TARGET>``

Usage example:  Promote local props changes (made via the UI) to the 'default' folder

.. code-block:: sh

    ksconf local/props.conf default/props.conf

"""

from __future__ import absolute_import, unicode_literals

import argparse
import os
import shutil
from copy import deepcopy

from ksconf.command import ConfDirProxy, ConfFileType, KsconfCmd, dedent
from ksconf.conf.delta import (DIFF_OP_DELETE, DIFF_OP_EQUAL, DIFF_OP_INSERT,
                               DIFF_OP_REPLACE, DiffStanza, DiffStzKey,
                               compare_cfgs, show_diff, summarize_cfg_diffs)
from ksconf.conf.merge import merge_conf_dicts
from ksconf.conf.parser import PARSECONF_STRICT, PARSECONF_STRICT_NC
from ksconf.consts import (EXIT_CODE_EXTERNAL_FILE_EDIT,
                           EXIT_CODE_FAILED_SAFETY_CHECK,
                           EXIT_CODE_NOTHING_TO_DO, EXIT_CODE_USER_QUIT)
from ksconf.filter import FilteredList, create_filtered_list
from ksconf.util.completers import conf_files_completer
from ksconf.util.file import _samefile, file_fingerprint

""" Possible behaviors.... thinking through what CLI options make the most sense...

 Things we may want to control:

     Q: What mode of operation?
         1.)  Automatic (merge all)
         2.)  Interactive (user guided / sub-shell)
         3.)  Batch mode:  CLI driven based on a stanza or key using either a name or pattern to
              select which content should be integrated.

     Q: What happens to the original?
         1.)  Updated
           a.)  Only remove source content that has been integrated into the target.
           b.)  Let the user pick
         2.)  Preserved  (Dry-run, or don't delete the original mode);  if output is stdout.
         3.)  Remove
           a.)  Only if all content was integrated.
           b.)  If user chose to discard entry.
           c.)  Always (--always-remove)
     Q: What to do with discarded content?
         1.)  Remove from the original (destructive)
         2.)  Place in a "discard" file.  (Allow the user to select the location of the file.)
         3.)  Automatically backup discards to a internal store, and/or log.  (More difficult to
              recover, but content is always logged/recoverable with some effort.)


 Interactive approach:

     3 action options:
         Integrate/Accept: Move content from the source to the target  (e.g., local to default)
         Reject/Remove:    Discard content from the source; destructive (e.g., rm local setting)
         Skip/Keep:        Don't push to target or remove from source (no change)

"""


def empty_dict(d):
    # Or just   d == {}?   Not sure which is better
    return isinstance(d, dict) and len(d) == 0


class PromoteCmd(KsconfCmd):
    help = dedent("""\
    Promote .conf settings between layers using either batch or interactive mode.

    Frequently this is used to promote conf changes made via the UI (stored in
    the ``local`` folder) to a version-controlled directory, such as ``default``.
    """)
    description = dedent("""\
    Propagate .conf settings applied in one file to another.  Typically this is used
    to move ``local`` changes (made via the UI) into another layer, such as the
    ``default`` or a named ``default.d/50-xxxxx``) folder.

    Promote has two modes:  batch and interactive.  In batch mode, all changes are
    applied automatically and the (now empty) source file is removed.  In interactive
    mode, the user is prompted to select stanzas to promote.  This way local changes
    can be held without being promoted.

    NOTE: Changes are *MOVED* not copied, unless ``--keep`` is used.
    """)
    format = "manual"
    maturity = "beta"

    def register_args(self, parser: argparse.ArgumentParser):
        parser.set_defaults(mode="ask")
        parser.add_argument("source", metavar="SOURCE",
                            type=ConfFileType("r+", "load", parse_profile=PARSECONF_STRICT_NC),
                            help="The source configuration file to pull changes from. "
                                 "(Typically the :file:`local` conf file)"
                            ).completer = conf_files_completer
        parser.add_argument("target", metavar="TARGET",
                            type=ConfFileType("r+", "none", accept_dir=True,
                                              parse_profile=PARSECONF_STRICT), help=dedent("""\
            Configuration file or directory to push the changes into.
            (Typically the :file:`default` folder)
            """)
                            ).completer = conf_files_completer
        grp1 = parser.add_mutually_exclusive_group()
        grp1.add_argument("--batch", "-b", action="store_const",
                          dest="mode", const="batch", help=dedent("""\
            Use batch mode where all configuration settings are automatically promoted.
            All changes are removed from source and applied to target.
            The source file will be removed unless
            ``--keep-empty`` is used."""))
        grp1.add_argument("--interactive", "-i",
                          action="store_const",
                          dest="mode", const="interactive", help=dedent("""\
            Enable interactive mode where the user will be prompted to approve
            the promotion of specific stanzas and attributes.
            The user will be able to apply, skip, or edit the changes being promoted."""))
        grp1.add_argument("--summary", "-s",
                          action="store_const",
                          dest="mode", const="summary",
                          help="Summarize content that could be promoted.")
        grp1.add_argument("--diff", "-d",
                          action="store_const",
                          dest="mode", const="diff",
                          help="Show the diff of what would be promoted.")

        parser.add_argument("--verbose", action="store_true", default=False,
                            help="Enable additional output.")

        pg_ftr = parser.add_argument_group("Automatic filtering options", dedent("""\
            Include or exclude stanzas to promote using these filter options.
            Stanzas selected by these filters will be promoted.

            All filter options can be provided multiple times.
            If you have a long list of filters, they can be saved in a file and
            referenced using the special ``file://`` prefix.  One entry per line."""))
        pg_ftr.add_argument("--match", "-m",
                            choices=["regex", "wildcard", "string"],
                            default="wildcard",
                            help=dedent("""\
            Specify pattern matching mode.
            Defaults to 'wildcard' allowing for ``*`` and  ``?`` matching.
            Use 'regex' for more power but watch out for shell escaping.
            Use 'string' to enable literal matching."""))
        pg_ftr.add_argument("--ignore-case", action="store_true",
                            help=dedent("""\
            Ignore case when comparing or matching strings.
            By default matches are case-sensitive."""))
        pg_ftr.add_argument("--invert-match", "-v", action="store_true",
                            help=dedent("""\
            Invert match results.
            This can be used to prevent content from being promoted."""))
        pg_ftr.add_argument("--stanza", metavar="PATTERN", action="append", default=[],
                            help=dedent("""\
            Promote any stanza with a name matching the given pattern.
            PATTERN supports bulk patterns via the ``file://`` prefix."""))

        parser.add_argument("--force", "-f",
                            action="store_true", default=False,
                            help="Disable safety checks. "
                            "Don't check to see if SOURCE and TARGET share the same basename.")
        parser.add_argument("--keep", "-k",
                            action="store_true", default=False, help=dedent("""\
            Keep conf settings in the source file.
            All changes will be copied into the TARGET file instead of being moved there.
            This is typically a bad idea since local always overrides default."""))
        parser.add_argument("--keep-empty",
                            action="store_true", default=False, help=dedent("""\
            Keep the source file, even if after the settings promotions the file has no content.
            By default, SOURCE will be removed after all content has been moved into TARGET.
            Splunk will re-create any necessary local files on the fly."""))

    def run(self, args):
        if isinstance(args.target, ConfDirProxy):
            # If a directory is given instead of a target file, then assume the source filename
            # and target filename are the same.
            # Also handle local/default meta:     e.g.:   ksconf promote local.meta .
            source_basename = os.path.basename(args.source.name)
            if source_basename == "local.meta":
                args.target = args.target.get_file("default.meta")
            else:
                args.target = args.target.get_file(source_basename)
            del source_basename

        if not os.path.isfile(args.target.name):
            self.stdout.write(f"Target file {args.target.name} does not exist.  "
                              f"Moving source file {args.source.name} to the target.\n")
            # For windows:  Close out any open file descriptors first
            args.target.close()
            args.source.close()
            if args.keep:
                shutil.copy2(args.source.name, args.target.name)
            else:
                shutil.move(args.source.name, args.target.name)
            return

        # If src/dest are the same, then the file ends up being deleted.  Whoops!
        if _samefile(args.source.name, args.target.name):
            self.stderr.write("Aborting.  SOURCE and TARGET are the same file!\n")
            return EXIT_CODE_FAILED_SAFETY_CHECK

        fp_source = file_fingerprint(args.source.name)
        fp_target = file_fingerprint(args.target.name)

        # Todo: Add a safety check prevent accidental merge of unrelated files.
        # Scenario: promote local/props.conf into default/transforms.conf
        # Possible check (1) Are basenames are different?  (props.conf vs transforms.conf)
        # Possible check (2) Are there key's in common? (DEST_KEY vs REPORT)
        # Using #1 for now, consider if there's value in #2
        bn_source = os.path.basename(args.source.name)
        bn_target = os.path.basename(args.target.name)
        if bn_source.endswith(".meta") and bn_target.endswith(".meta"):
            # Allow local.meta -> default.meta without --force or a warning message
            pass
        elif bn_source != bn_target:
            if args.mode in ("summary", "diff"):
                pass
            # Todo: Allow for interactive prompting when in interactive but not force mode.
            elif args.force:
                self.stderr.write(
                    f"Promoting content across conf file types ({bn_source} --> {bn_target}) because the "
                    "'--force' CLI option was set.\n")
            else:
                self.stderr.write(
                    "Refusing to promote content between different types of configuration "
                    f"files.  {bn_source} --> {bn_target}  If this is intentional, override this safety "
                    "check with '--force'\n")
                return EXIT_CODE_FAILED_SAFETY_CHECK

        # Todo:  Preserve comments in the TARGET file.  Worry with promoting of comments later...
        # Parse all config files
        cfg_src = args.source.data
        cfg_tgt = args.target.data

        if not cfg_src:
            self.stderr.write(f"No settings in {args.source.name}.  Nothing to promote.\n")
            return EXIT_CODE_NOTHING_TO_DO

        # Prep filters and determine if there's any automatic work that can be done
        if self.prep_filters(args):
            # Run filter and then return control back
            delta, cfg_src, cfg_tgt = self._do_promote_list(cfg_src, cfg_tgt, args)
        else:
            delta = compare_cfgs(cfg_tgt, cfg_src, replace_level="key")
            delta = [op for op in delta if op.tag != DIFF_OP_DELETE]

        if args.mode == "diff":
            self.stderr.write("\n")
            show_diff(self.stderr, delta)
            return

        if args.mode in ("ask", "summary"):
            # Show a summary of how many new stanzas would be copied across; how many key changes.
            # And either accept all (batch) or pick selectively (batch)
            summarize_cfg_diffs(delta, self.stderr)
            if args.mode == "summary":
                return

            while True:
                resp = input("Would you like to apply ALL changes?  (y/n/d/q) ")
                resp = resp[:1].lower()
                if resp == 'q':
                    return EXIT_CODE_USER_QUIT
                elif resp == 'd':
                    show_diff(self.stdout, delta, headers=(args.source.name, args.target.name))
                elif resp == 'y':
                    args.mode = "batch"
                    break
                elif resp == 'n':
                    args.mode = "interactive"
                    break

        if args.mode == "interactive":
            (cfg_final_src, cfg_final_tgt) = self._do_promote_interactive(cfg_src, cfg_tgt, args)
        elif args.stanza:
            # We already applied changes above.
            cfg_final_src, cfg_final_tgt = cfg_src, cfg_tgt
        else:
            (cfg_final_src, cfg_final_tgt) = self._do_promote_automatic(cfg_src, cfg_tgt, args)

        # Minimize race condition:  Do file mtime/hash check here.  Abort on external change.
        # Todo: Eventually use temporary files and atomic renames to further minimize the risk
        # Todo: Make backup '.bak' files (user configurable)
        # Todo: Avoid rewriting files if NO changes were made. (preserve prior backups)
        # Todo: Restore file modes and such

        if file_fingerprint(args.source.name, fp_source):
            self.stderr.write("Aborting!  External source file changed: {0}\n".
                              format(args.source.name))
            return EXIT_CODE_EXTERNAL_FILE_EDIT
        if file_fingerprint(args.target.name, fp_target):
            self.stderr.write("Aborting!  External target file changed: {0}\n".
                              format(args.target.name))
            return EXIT_CODE_EXTERNAL_FILE_EDIT
        # Reminder:  conf entries are being removed from source and promoted into target
        args.target.dump(cfg_final_tgt)
        if not args.keep:
            # If --keep is set, we never touch the source file.
            if cfg_final_src:
                args.source.dump(cfg_final_src)
            else:
                # Config file is empty.  Should we write an empty file, or remove it?
                if args.keep_empty:
                    args.source.dump(cfg_final_src)
                else:
                    args.source.unlink()

    def prep_filters(self, args):
        flags = 0
        if args.ignore_case:
            flags |= FilteredList.IGNORECASE
        if args.verbose:
            flags |= FilteredList.VERBOSE
        self.stanza_filters = create_filtered_list(args.match, flags).feedall(args.stanza)
        if args.stanza:
            return True

    def apply_filters(self, delta, invert_match=False):
        for op in delta:
            if self.stanza_filters.match_stanza(op.location.stanza) ^ invert_match:
                yield op

    @staticmethod
    def combine_stanza(a, b):
        if a is None:
            return b
        d = dict(a)
        d.update(b)
        return d

    def _do_promote_automatic(self, cfg_src, cfg_tgt, args):
        # Promote ALL entries;  simply, isn't it...  ;-)
        final_cfg = merge_conf_dicts(cfg_tgt, cfg_src)
        return ({}, final_cfg)

    def _do_promote_interactive(self, cfg_src, cfg_tgt, args):
        ''' Interactively "promote" settings from one configuration file into another

        Model after git's "patch" mode, from git docs:

        This lets you choose one path out of a status like selection. After choosing the path, it
        presents the diff between the index and the working tree file and asks you if you want to
        stage the change of each hunk. You can select one of the following options and type return:

           y - stage this hunk
           n - do not stage this hunk
           q - quit; do not stage this hunk or any of the remaining ones
           a - stage this hunk and all later hunks in the file
           d - do not stage this hunk or any of the later hunks in the file
           g - select a hunk to go to
           / - search for a hunk matching the given regex
           j - leave this hunk undecided, see next undecided hunk
           J - leave this hunk undecided, see next hunk
           k - leave this hunk undecided, see previous undecided hunk
           K - leave this hunk undecided, see previous hunk
           s - split the current hunk into smaller hunks
           e - manually edit the current hunk
           ? - print help


        Note:  In git's "edit" mode you are literally editing a patch file, so you can modify both
        the working tree file as well as the file that's being staged.  While this is nifty, as
        git's own documentation points out (in other places), that "some changes may have confusing
        results".  Therefore, it probably makes sense to limit what the user can edit.

        ============================================================================================

        Options we may be able to support:

           Pri k   Description
           --- -   -----------
           [1] y - stage this section or key
           [1] n - do not stage this section or key
           [1] q - quit; do not stage this or any of the remaining sections or attributes
           [2] a - stage this section or key and all later sections in the file
           [2] d - do not stage this section or key or any of the later section or key in the file
           [1] s - split the section into individual attributes
           [3] e - edit the current section or key
           [2] ? - print help

        Q:  Is it less confusing to the user to adopt the 'local' and 'default' paradigm here?
        Even though we know that change promotions will not *always* be between default and local.
        (We can and should assume some familiarity with Splunk conf terms, less so than familiarity
        with git lingo.)
        '''

        def prompt_yes_no(prompt):
            print("")
            while True:
                r = input(prompt + " (y/n) ")
                if r.lower().startswith("y"):
                    return True
                elif r.lower().startswith("n"):
                    return False

        out_src = deepcopy(cfg_src)
        out_cfg = deepcopy(cfg_tgt)
        # Todo:  IMPLEMENT A MANUAL MERGE/DIFF HERE:
        # What ever is migrated, move it OUT of cfg_src, and into cfg_tgt

        diff = compare_cfgs(cfg_tgt, cfg_src, replace_level="key")
        for op in diff:
            if op.tag == DIFF_OP_DELETE:
                # This is normal.   Only changed attributes are copied & updated in local.
                continue
            elif op.tag == DIFF_OP_EQUAL:
                # Q:  Should we simply remove everything from the source file that already lines
                #     up with the target?  Just ask
                if isinstance(op.location, DiffStzKey):
                    msg = f"[{op.location.stanza}]  {op.location.key}"
                elif isinstance(op.location, DiffStanza):
                    msg = f"[{op.location.stanza}]"
                if prompt_yes_no(f"Remove matching entry {msg}  "):
                    if isinstance(op.location, DiffStanza):
                        del out_src[op.location.stanza]
                    else:
                        del out_src[op.location.stanza][op.location.key]
            else:
                '''
                self.stderr.write("Found change:  <{0}> {1!r}\n-{2!r}\n+{3!r}\n\n\n"
                    .format(op.tag, op.location, op.b, op.a))
                '''
                if isinstance(op.location, DiffStanza):
                    # Move entire stanza
                    """ ----  If we need to support empty stanza promotion.....
                    print(f"OP:   {op!r}")
                    if empty_dict(op.b):
                        print(f"Empty stanza [{op.location.stanza}]")
                    else:
                    """
                    if True:
                        show_diff(self.stdout, [op])
                    if prompt_yes_no(f"Apply [{op.location.stanza}]"):
                        out_cfg[op.location.stanza] = self.combine_stanza(op.a, op.b)
                        del out_src[op.location.stanza]
                else:
                    show_diff(self.stdout, [op])
                    # Logically, this must be either an INSERT or REPLACE
                    if prompt_yes_no(f"Apply [{op.location.stanza}] {op.location.key}"):
                        # Move key
                        out_cfg[op.location.stanza][op.location.key] = op.b
                        del out_src[op.location.stanza][op.location.key]
                        # If last remaining key in the src stanza?  Then delete the entire stanza
                        if not out_src[op.location.stanza]:
                            del out_src[op.location.stanza]
        return (out_src, out_cfg)

    def _do_promote_list(self, cfg_src, cfg_tgt, args):
        out_src = deepcopy(cfg_src)
        out_cfg = deepcopy(cfg_tgt)
        diff = compare_cfgs(cfg_tgt, cfg_src, replace_level="key")
        diff = [op for op in self.apply_filters(diff, args.invert_match)
                if op.tag in (DIFF_OP_INSERT, DIFF_OP_REPLACE)]
        for op in diff:
            if args.verbose:
                show_diff(self.stdout, [op])
            if isinstance(op.location, DiffStanza):
                # Move entire stanza
                out_cfg[op.location.stanza] = self.combine_stanza(op.a, op.b)
                del out_src[op.location.stanza]
            else:
                # Move key
                out_cfg[op.location.stanza][op.location.key] = op.b
                del out_src[op.location.stanza][op.location.key]
                # If last remaining key in the src stanza?  Then delete the entire stanza
                if not out_src[op.location.stanza]:
                    del out_src[op.location.stanza]
        return (diff, out_src, out_cfg)



from __future__ import absolute_import, unicode_literals

import fnmatch
import os
import re
import sys
from collections import Counter
from pathlib import Path

from ksconf.conf.parser import GLOBAL_STANZA
from ksconf.util.file import splglob_to_regex

# Note on future direction:
#
# It seems like each item on the list should be capable of having it's own type and flags (which
# could be inherited, at may not be known at the time the rules are first initialized).
#
# There should still be a phase that each "Rule" goes though to (1) set core attributes, (2) prep
# things that can be compiled, syntax checked, and any default flags should be available by this
# time, and (3) evaluate against a specific values for true/false.  This would allow things like a
# special prefix to lazy-switch modes (for example: ~regex~, or only do fnmatching if there's a
# wildcard otherwise stick with simple string matching, for efficiency...)


class FilteredList:
    IGNORECASE = 1
    INVERT = 2
    VERBOSE = 4

    def __init__(self, flags=0, default=True):
        self.data = []
        self.rules = None
        self.counter = Counter()
        self.flags = flags
        self._prep = True
        #  If no patterns defined, return default.  (True => match everything)
        self.default = default

    def _feed_from_file(self, path):
        items = []
        with open(path) as f:
            for line in f:
                line = line.rstrip()
                # Skip empty or "comment" lines
                if line and line[0] != "#":
                    items.append(line)
        if self.flags & self.VERBOSE:
            sys.stderr.write(f"Loaded {len(items)} patterns from {path}\n")
        return items

    def feed(self, item, filter=None):
        if isinstance(item, Path):
            item = os.fspath(item)

        if item.startswith("file://"):
            # File ingestion mode
            filename = item[7:]
            for item in self._feed_from_file(filename):
                self.feed(item, filter)
        else:
            if filter:
                item = filter(item)
            self.data.append(item)
        # New items added.  Mark prep-work as incomplete
        self._prep = False

    def feedall(self, iterable, filter=None):
        if iterable:
            for i in iterable:
                self.feed(i, filter)
        return self

    def _pre_match(self):  # pragma: no cover
        pass

    def match(self, item):
        if self.data:
            # Kick off any first-time preparatory activities
            if self._prep is False:
                self._pre_match()
                self.reset_counters()
                self._prep = True

            ret = self._match(item)
            if ret:
                self.counter[ret] += 1
                result = True
            else:
                result = False

        else:
            #  No patterns defined, use default
            return self.default
        if self.flags & self.INVERT:
            return not result
        else:
            return result

    def match_path(self, path):
        if isinstance(path, Path):
            path = os.fspath(path)
        if os.path.sep != "/":
            path = path.replace(os.path.sep, "/")
        return self.match(path)

    def match_stanza(self, stanza):
        """ Same as match(), but handle GLOBAL_STANZA gracefully. """
        if stanza is GLOBAL_STANZA:
            stanza = "default"
        return self.match(stanza)

    def reset_counters(self):
        # Set all the counters to 0, so the caller can know which filters had 0 hits
        self.counter = Counter()
        self.counter.update((n, 0) for n in self.data)

    @property
    def has_rules(self):
        return bool(self.data)

    def _match(self, item):  # pragma: no cover
        """ Return name of rule, indicating a match or not. """
        raise NotImplementedError


class FilteredListString(FilteredList):
    """ Handle simple string comparisons """

    def _pre_match(self):
        if self.flags & self.IGNORECASE:
            # Lower-case all strings in self.data.  (Only need to do this once)
            self.rules = {i.lower() for i in self.data}
        else:
            self.rules = set(self.data)
        return self.rules

    def _match(self, item):
        if self.flags & self.IGNORECASE:
            item = item.lower()
        if item in self.rules:
            return item
        else:
            return False

    def reset_counters(self):
        self.counter = Counter()
        self.counter.update({n: 0 for n in self.rules})


class FilteredListRegex(FilteredList):
    """ Regular Expression support """

    def calc_regex_flags(self):
        re_flags = 0
        if self.flags & self.IGNORECASE:
            re_flags |= re.IGNORECASE
        return re_flags

    def _pre_match(self):
        # Compile all regular expressions
        re_flags = self.calc_regex_flags()
        # XXX: Add better error handling here for friendlier user feedback
        self.rules = [(pattern, re.compile(pattern, re_flags)) for pattern in self.data]

    def _match(self, item):
        for name, pattern_re in self.rules:
            if pattern_re.match(item):
                return name
        return False

    def reset_counters(self):
        self.counter = Counter()
        self.counter.update({i[0]: 0 for i in self.rules})


class FilteredListWildcard(FilteredListRegex):
    """ Wildcard support (handling '*' and ?')
    Technically fnmatch also supports [] and [!] character ranges, but we don't advertise that
    """

    def _pre_match(self):
        # Use fnmatch to translate wildcard expression to a regex, and compile regex
        re_flags = self.calc_regex_flags()
        self.rules = [(wc, re.compile(fnmatch.translate(wc), re_flags)) for wc in self.data]


class FilteredListSplunkGlob(FilteredListRegex):
    """ Classic wildcard support ('*' and ?') plus '...' or '**' for multiple-path components with
    some (non-advertised) pass-through regex behavior
    """

    def _pre_match(self):
        # Use splglob_to_regex to translate wildcard expression to a regex, and compile regex
        re_flags = self.calc_regex_flags()
        self.rules = [(wc, splglob_to_regex(wc, re_flags)) for wc in self.data]


class_mapping = {
    "string": FilteredListString,
    "wildcard": FilteredListWildcard,
    "regex": FilteredListRegex,
    "splunk": FilteredListSplunkGlob,
}


def create_filtered_list(match_mode: str, flags=0, default=True) -> FilteredList:
    try:
        class_ = class_mapping[match_mode]
    except KeyError:
        raise NotImplementedError(f"Matching mode {match_mode!r} undefined")
    return class_(flags, default)



""" SUBCOMMAND:  ``ksconf sort <CONF>``

Usage example:  To recursively sort all files (in-place):

.. code-block:: sh

    find . -name '*.conf' | xargs ksconf sort -i

"""
from __future__ import absolute_import, unicode_literals

from ksconf.command import KsconfCmd, dedent
from ksconf.conf.parser import PARSECONF_STRICT, ConfParserException, write_conf
from ksconf.consts import (EXIT_CODE_BAD_CONF_FILE, EXIT_CODE_SORT_APPLIED,
                           EXIT_CODE_SUCCESS, SMART_NOCHANGE)
from ksconf.util.completers import conf_files_completer
from ksconf.util.file import expand_glob_list


def _has_nosort_marker(path):
    # KISS:  Look for the KSCONF-NO-SORT string in the first 4k of this file.
    with open(path, "rb") as stream:
        prefix = stream.read(4096)
    return b"KSCONF-NO-SORT" in prefix


class SortCmd(KsconfCmd):
    help = "Sort a Splunk .conf file creating a normalized format appropriate for version control"
    description = dedent("""\
    Sort a Splunk .conf file.  Sort has two modes:  (1) by default, the sorted
    config file will be echoed to the screen.  (2) the config files are updated
    in-place when the ``-i`` option is used.

    Manually managed conf files can be protected against changes by adding a comment containing the
    string ``KSCONF-NO-SORT`` to the top of any .conf file.
    """)
    format = "manual"
    maturity = "stable"

    def register_args(self, parser):
        import argparse
        parser.add_argument("conf", metavar="FILE", nargs="+",
                            default=["-"],
                            help="Input file to sort, or standard input."
                            ).completer = conf_files_completer

        # Pick mode:  target (sysout) vs inplace
        mode = parser.add_mutually_exclusive_group()
        mode.add_argument("--target", "-t", metavar="FILE",
                          type=argparse.FileType('w'), default=self.stdout,
                          help="File to write results to.  Defaults to standard output."
                          ).completer = conf_files_completer
        mode.add_argument("--inplace", "-i",
                          action="store_true", default=False, help=dedent("""\
                          Replace the input file with a sorted version.

                          WARNING:  This a potentially destructive operation that
                          may move/remove comments."""))

        # Inplace update arguments
        grp1 = parser.add_argument_group("In-place update arguments")
        grp1.add_argument("-F", "--force", action="store_true",
                          help=dedent("""\
                          Force file sorting for all files, even for files containing the special
                          'KSCONF-NO-SORT' marker."""))
        grp1.add_argument("-q", "--quiet", action="store_true",
                          help=dedent("""\
                          Reduce the output.
                          Reports only updated or invalid files.
                          This is useful for pre-commit hooks, for example."""))

        parser.add_argument("-n", "--newlines", metavar="LINES", type=int, default=1,
                            help="Number of lines between stanzas.")

    def pre_run(self, args):
        # For Windows users, expand any glob patterns as needed.
        args.conf = list(expand_glob_list(args.conf))

    def run(self, args):
        ''' Sort one or more configuration file. '''
        stanza_delims = "\n" * args.newlines
        self.parse_profile = PARSECONF_STRICT
        if args.inplace:
            failure = False
            changes = 0
            for conf in args.conf:
                try:
                    if not args.force and _has_nosort_marker(conf):
                        if not args.quiet:
                            self.stderr.write(f"Skipping no-sort file {conf}\n")
                        continue
                    c = self.parse_conf(conf, mode='r+', raw_exec=True)
                    # c = parse_conf(conf, profile=PARSECONF_STRICT)
                    data = c.data
                    smart_rc = c.dump(c.data, stanza_delim=stanza_delims, sort=True)
                    # smart_rc = smart_write_conf(conf, data, stanza_delim=stanza_delims,
                    #                            sort=True)
                except ConfParserException as e:
                    smart_rc = None
                    self.stderr.write(f"Error trying to process file {conf}.  "
                                      f"Error:  {e}\n")
                    failure = True
                    continue
                if smart_rc == SMART_NOCHANGE:
                    if not args.quiet:
                        self.stderr.write("Nothing to update.  "
                                          f"File {conf} is already sorted\n")
                else:
                    self.stderr.write(f"Replaced file {conf} with sorted content.\n")
                    changes += 1
            if failure:
                return EXIT_CODE_BAD_CONF_FILE
            if changes:
                return EXIT_CODE_SORT_APPLIED
        else:
            for conf in args.conf:
                if len(args.conf) > 1:
                    args.target.write(f"---------------- [ {conf} ] ----------------\n\n")
                data = self.parse_conf(conf).data
                write_conf(args.target, data, stanza_delim=stanza_delims, sort=True)
            return EXIT_CODE_SUCCESS



from __future__ import unicode_literals

from typing import IO, List, Set, Tuple, TypeVar

from ksconf.types import PathType


def fileobj_compare(f1: IO, f2: IO) -> bool:
    # Borrowed from filecmp
    f1.seek(0)
    f2.seek(0)
    buffsize = 8192
    while True:
        b1 = f1.read(buffsize)
        b2 = f2.read(buffsize)
        if b1 != b2:
            return False
        if not b1:
            return True


def file_compare(fn1: PathType, fn2: PathType) -> bool:
    with open(fn1, "rb") as f1, \
            open(fn2, "rb") as f2:
        return fileobj_compare(f1, f2)


T = TypeVar("T")


def cmp_sets(a: Set[T], b: Set[T]) -> Tuple[List[T], List[T], List[T]]:
    """ Result tuples in format (a-only, common, b-only) """
    set_a = set(a)
    set_b = set(b)
    a_only = sorted(set_a.difference(set_b))
    common = sorted(set_a.intersection(set_b))
    b_only = sorted(set_b.difference(set_a))
    return (a_only, common, b_only)


# For backwards compatibility
_cmp_sets = cmp_sets



from __future__ import absolute_import, unicode_literals

import os
from fnmatch import fnmatch
from typing import ByteString, Callable, Iterable, NamedTuple, Optional, Sequence, Tuple

from ksconf.consts import RegexType


class GenArchFile(NamedTuple):
    path: str
    mode: int
    size: int
    payload: Optional[ByteString]


def extract_archive(archive_name, extract_filter: Optional[Callable] = None) -> Iterable[GenArchFile]:
    if extract_filter is not None and not callable(extract_filter):  # pragma: no cover
        raise ValueError("extract_filter must be a callable!")
    archive_name = os.fspath(archive_name)
    if archive_name.lower().endswith(".zip"):
        return _extract_zip(archive_name, extract_filter)
    else:
        return _extract_tar(archive_name, extract_filter)


def gaf_filter_name_like(pattern):
    def filter(gaf):
        filename = os.path.basename(gaf.path)
        return fnmatch(filename, pattern)

    return filter


def _extract_tar(path, extract_filter=None, encoding="utf-8"):
    import tarfile
    with tarfile.open(path, "r", encoding=encoding) as tar:
        for ti in tar:
            if not ti.isreg():
                '''
                print(f"Skipping {ti.name}  ({ti.type})")
                '''
                continue
            mode = ti.mode & 0o777
            if extract_filter is None or \
                    extract_filter(GenArchFile(ti.name, mode, ti.size, None)):
                tar_file_fp = tar.extractfile(ti)
                buf = tar_file_fp.read()
            else:
                buf = None
            yield GenArchFile(ti.name, mode, ti.size, buf)


def _extract_zip(path, extract_filter=None, mode=0o644, encoding="latin"):
    # Note:  No encoding defined for Zip file spec.  Sticking with what WinZip uses by default.
    import zipfile
    with zipfile.ZipFile(path, mode="r") as zipf:
        for zi in zipf.infolist():
            if hasattr(zi.filename, "decode"):
                zi.filename = zi.filename.decode(encoding)
            if zi.filename.endswith('/'):
                # Skip directories
                continue
            if extract_filter is None or \
                    extract_filter(GenArchFile(zi.filename, mode, zi.file_size, None)):
                payload = zipf.read(zi)
            else:
                payload = None
            yield GenArchFile(zi.filename, mode, zi.file_size, payload)


def sanity_checker(iterable: Iterable[GenArchFile]) -> Iterable[GenArchFile]:
    # Keep this here for a few versions because some of the cdillc.splunk code references this.
    from warnings import warn
    warn("Please use AppManifest.check_paths() instead.  The sanity_checker() "
         "function will be removed in 1.0 or sooner.", DeprecationWarning)
    for gaf in iterable:
        if gaf.path.startswith("/") or ".." in gaf.path:
            raise ValueError(f"Bad path found in archive:  {gaf.path}")
        yield gaf


def gen_arch_file_remapper(iterable: Iterable[GenArchFile],
                           mapping: Sequence[Tuple[str, str]]
                           ) -> Iterable[GenArchFile]:
    # Mapping is assumed to be a sequence of (find,replace) strings; find can be compiled regex
    for gaf in iterable:
        path = gaf.path
        for (find, replace) in mapping:
            if isinstance(find, RegexType):
                path = find.sub(replace, path)
            else:
                path = path.replace(find, replace)
        if gaf.path == path:
            yield gaf
        else:
            yield GenArchFile(path, gaf.mode, gaf.size, gaf.payload)



from __future__ import absolute_import, unicode_literals

# Optional argcomplete library for CLI (BASH-based) tab completion
try:
    from argcomplete import autocomplete
    from argcomplete.completers import DirectoriesCompleter, FilesCompleter
except ImportError:

    def _argcomplete_noop(*args, **kwargs):
        del args, kwargs

    autocomplete = _argcomplete_noop
    # noinspection PyPep8Naming
    FilesCompleter = DirectoriesCompleter = _argcomplete_noop

# Someday add *.meta (once more testing is done with those files
conf_files_completer = FilesCompleter(allowednames=["*.conf"])



"""r SUBCOMMAND:  ``ksconf attr-get <CONF> --stanza STANZA --attribute ATTR``

.. code-block:: sh

    ksconf attr-get $SPLUNK_HOME/etc/apps/Splunk_TA_aws/default/app.conf --stanza launcher --attribute version


SUBCOMMAND:  ``ksconf attr-set <CONF> --stanza STANZA --attribute ATTR --value VALUE``

.. code-block:: sh

    ksconf attr-set $SPLUNK_HOME/etc/apps/Splunk_TA_aws/local/app.conf \
        --stanza launcher --attribute version --value 9.9.9

    echo "9.9.9" > /tmp/new_version
    ksconf attr-set $SPLUNK_HOME/etc/apps/Splunk_TA_aws/local/app.conf \
        --stanza launcher --attribute version  -t file /tmp/new_version

    export NEW_VERSION=1.2.3
    ksconf attr-set $SPLUNK_HOME/etc/apps/Splunk_TA_aws/local/app.conf \
        --stanza launcher --attribute version -t env NEW_VERSION

"""

from __future__ import absolute_import, unicode_literals

import argparse
import os
from pathlib import Path

from ksconf.command import KsconfCmd, dedent
from ksconf.conf.parser import update_conf
from ksconf.consts import (EXIT_CODE_CONF_NO_DATA_MATCH, EXIT_CODE_NO_SUCH_FILE,
                           EXIT_CODE_NOTHING_TO_DO, EXIT_CODE_SUCCESS)
from ksconf.util.completers import conf_files_completer
from ksconf.util.file import expand_glob_list


class AttrGetCmd(KsconfCmd):
    help = "Get the value from a specific stanzas and attribute"
    description = dedent("""\
    Get a specific stanza and attribute value from a Splunk .conf file.
    """)
    format = "manual"
    maturity = "beta"

    def register_args(self, parser):
        parser.add_argument("conf", nargs="+",
                            default=["-"],
                            help="Input file or standard input."
                            ).completer = conf_files_completer

        parser.add_argument("--stanza", "-s", required=True,
                            help="Name of the stanza within CONF to retrieve.")
        parser.add_argument("--attribute", "--attr", "-a",
                            metavar="ATTR", required=True,
                            help="Name of attribute within STANZA to retrieve.")

        parser.add_argument("--missing-okay", action="store_true", default=False,
                            help="Ignore missing stanzas and attributes.  ")
        '''
        parser.add_argument("--quite", "-q", action="store_true", default=False,
                            help="Reduce the amount of output stderr")
        '''

        parser.add_argument("-o", "--output",
                            type=argparse.FileType('w'), default=self.stdout,
                            help="File where the filtered results are written.  "
                                 "Defaults to standard out.")

    def pre_run(self, args):
        # For Windows users, expand any glob patterns as needed.
        args.conf = list(expand_glob_list(args.conf))

    def run(self, args):
        ''' For a given conf file, get the 'value' from [stanza] attribute = value '''
        for conf in args.conf:
            if len(args.conf) > 1:
                args.output.write(f"---------------- [ {conf} ] ----------------\n\n")
            data = self.parse_conf(conf).data
            if args.missing_okay:
                try:
                    value = data[args.stanza][args.attribute]
                except KeyError:
                    value = ""
            else:
                try:
                    stanza = data[args.stanza]
                except KeyError:
                    self.stderr.write(f"File {conf} does not have the stanza [{args.stanza}] \n")
                    return EXIT_CODE_CONF_NO_DATA_MATCH
                try:
                    value = stanza[args.attribute]
                except KeyError:
                    self.stderr.write(f"File {conf} does not have {args.attribute} "
                                      f"in stanza [{args.stanza}]\n")
                    return EXIT_CODE_CONF_NO_DATA_MATCH
            args.output.write(f"{value}\n")
            args.output.flush()

        return EXIT_CODE_SUCCESS


class AttrSetCmd(KsconfCmd):
    help = "Set the value of a specific stanzas and attribute"
    description = dedent("""\
    Set a specific stanza and attribute value of a Splunk .conf file.
    The value can be provided as a command line argument, file, or
    environment variable

    This command does not support preserving leading or trailing whitespace.
    Normally this is desireable.
    """)
    format = "manual"
    maturity = "beta"

    def register_args(self, parser):
        parser.add_argument("conf",
                            help="Configuration file to update."
                            ).completer = conf_files_completer
        parser.add_argument("--stanza", "-s", required=True,
                            help="Name of the stanza within CONF to set.")
        parser.add_argument("--attribute", "--attr", "-a",
                            metavar="ATTR", required=True,
                            help="Name of the attribute within STANZA to set.")

        parser.add_argument("value",
                            help="Value to apply to the conf file.  Note that this can be a raw "
                            "text string, or the name of the file, or an environment variable")

        parser.add_argument("--value-type", "-t", metavar="TYPE", default="string",
                            choices=["string", "file", "env"],
                            help="Select the type of VALUE.  The default is a string.  "
                            "Alternatively, the real value can be provided within a file, "
                            "or an environment variable.")

        parser.add_argument("--create-missing", action="store_true", default=False,
                            help="Create a new conf file if it doesn't currently exist.")
        parser.add_argument("--no-overwrite", action="store_true", default=False,
                            help="Only set VALUE if none currently exists.  "
                            "This can be used to safely set a one-time default, "
                            "but don't update overwrite an existing value.")

    def get_value(self, value, value_type):
        if value_type == "file":
            if value == "-":
                return self.stdin.read().strip()
            else:
                return Path(value).read_text().strip()
        elif value_type == "env":
            return os.environ[value]
        else:
            return value

    def set_conf_value(self, conf_file: Path, stanza: str, attribute: str, value: str,
                       create_missing: bool, no_overwrite: bool):

        if not conf_file.is_file() and not create_missing:
            self.stderr.write(f"Unable to write to non-existent file {conf_file}.\n"
                              "To automatically create missing files use '--create-missing'\n")
            return EXIT_CODE_NO_SUCH_FILE

        with update_conf(conf_file, make_missing=create_missing) as conf:
            try:
                existing_value = conf[stanza][attribute]
            except KeyError:
                existing_value = None

            if existing_value is not None:
                if existing_value == value:
                    self.stderr.write(f"No change necessary.  {conf_file} "
                                      f"[{stanza}] {attribute} already has desired value.\n")
                    conf.cancel()
                    return EXIT_CODE_SUCCESS

                if no_overwrite:
                    self.stderr.write(f"Skipping updating {conf_file} due to --no-overwrite.  "
                                      f"[{stanza}] {attribute} already set.\n")
                    conf.cancel()
                    return EXIT_CODE_NOTHING_TO_DO

            if stanza not in conf:
                conf[stanza] = {}
            conf[stanza][attribute] = value

        self.stderr.write(f"Set {conf_file} [{stanza}] {attribute} as requested.\n")
        return EXIT_CODE_SUCCESS

    def run(self, args):
        ''' For a given conf file, set [stanza] attribute = value '''
        value = self.get_value(args.value, args.value_type)
        conf_path = Path(args.conf)
        return self.set_conf_value(conf_path, args.stanza, args.attribute, value,
                                   args.create_missing, args.no_overwrite)



import datetime
import difflib
import os
from collections import Counter, defaultdict
from dataclasses import dataclass
from enum import Enum
from io import open
from os import PathLike, fspath
from typing import List, NamedTuple, Optional, Sequence, TextIO, Union

from ksconf.conf.parser import GLOBAL_STANZA, ConfType, StanzaType, _format_stanza, default_encoding
from ksconf.consts import EXIT_CODE_DIFF_CHANGE, EXIT_CODE_DIFF_EQUAL, EXIT_CODE_DIFF_NO_COMMON
from ksconf.util.compare import cmp_sets
from ksconf.util.terminal import ANSI_BOLD, ANSI_GREEN, ANSI_RED, ANSI_RESET, ANSI_YELLOW, TermColor

####################################################################################################
# DiffVerb logic


class DiffVerb(str, Enum):
    INSERT = "insert"
    DELETE = "delete"
    REPLACE = "replace"
    EQUAL = "equal"

    # Allow sorting
    def __gt__(self, other):
        return self.value > other.value

    def __lt__(self, other):
        return self.value < other.value


# Legacy names
DIFF_OP_INSERT = DiffVerb.INSERT
DIFF_OP_DELETE = DiffVerb.DELETE
DIFF_OP_REPLACE = DiffVerb.REPLACE
DIFF_OP_EQUAL = DiffVerb.EQUAL


class DiffLevel(str, Enum):
    GLOBAL = "global"
    STANZA = "stanza"
    KEY = "key"


class DiffGlobal(NamedTuple):
    type: DiffLevel


class DiffStanza(NamedTuple):
    type: DiffLevel
    stanza: str


class DiffStzKey(NamedTuple):
    type: DiffLevel
    stanza: str
    key: str


class DiffOp(NamedTuple):
    tag: DiffVerb
    location: Union[DiffGlobal, DiffStanza, DiffStzKey]
    a: Union[ConfType, StanzaType, str, None]
    b: Union[ConfType, StanzaType, str, None]


@dataclass
class DiffHeader:
    name: str
    mtime: Optional[float] = None

    def __init__(self, name, mtime=None):
        self.name = name
        if mtime:
            self.mtime = mtime
        else:
            self.detect_mtime()

    def detect_mtime(self):
        try:
            self.mtime = os.stat(self.name).st_mtime
        except OSError:
            self.mtime = 0

    def __str__(self):
        if isinstance(self.mtime, (int, float)):
            ts = datetime.datetime.fromtimestamp(self.mtime)
        else:
            ts = self.mtime
        return f"{fspath(self.name):50} {ts}"


def compare_stanzas(a: StanzaType, b: StanzaType,
                    stanza_name: str,
                    replace_level: DiffLevel = DiffLevel.GLOBAL
                    ) -> List[DiffOp]:
    """
    :param replace_level: If a and b have no common keys, is a single stanza-level
                      'replace' is issue unless ``replace_level="key"``
    :type replace_level: bool
    """
    if a == b:
        return [DiffOp(DiffVerb.EQUAL, DiffStanza(DiffLevel.STANZA, stanza_name), a, b)]
    elif b is None:
        # A only
        return [DiffOp(DiffVerb.DELETE, DiffStanza(DiffLevel.STANZA, stanza_name), a, None)]
    elif a is None:
        # B only
        return [DiffOp(DiffVerb.INSERT, DiffStanza(DiffLevel.STANZA, stanza_name), None, b)]
    else:
        return list(_compare_stanzas(a, b, stanza_name, replace_level))


def _compare_stanzas(a: StanzaType, b: StanzaType,
                     stanza_name: str,
                     replace_level: DiffLevel) -> List[DiffOp]:
    kv_a, kv_common, kv_b = cmp_sets(list(a.keys()), list(b.keys()))

    if replace_level in (DiffLevel.GLOBAL, DiffLevel.STANZA) and not kv_common:
        # No keys in common, just swap
        yield DiffOp(DiffVerb.REPLACE, DiffStanza(DiffLevel.STANZA, stanza_name), a, b)
        return

    # Level 2 - Key comparisons
    for key in kv_a:
        yield DiffOp(DIFF_OP_DELETE, DiffStzKey(DiffLevel.KEY, stanza_name, key), a[key], None)
    for key in kv_b:
        yield DiffOp(DIFF_OP_INSERT, DiffStzKey(DiffLevel.KEY, stanza_name, key), None, b[key])
    for key in kv_common:
        a_ = a[key]
        b_ = b[key]
        if a_ == b_:
            yield DiffOp(DiffVerb.EQUAL, DiffStzKey(DiffLevel.KEY, stanza_name, key), a_, b_)
        else:
            yield DiffOp(DiffVerb.REPLACE, DiffStzKey(DiffLevel.KEY, stanza_name, key), a_, b_)


def compare_cfgs(a: ConfType, b: ConfType,
                 replace_level: DiffLevel = DiffLevel.GLOBAL
                 ) -> List[DiffOp]:
    """
    Calculate a set of deltas which describes how to transform a into b.

    :param a: the first/original configuration entity
    :type a: dict
    :param b: the second/target configuration entity
    :type b: dict
    :param replace_level: The highest level 'replace' event that can be returned.
        Acceptable values are ``global``, ``stanza``, and ``key``.
        These examples may help:

            *   Using 'global' with identical inputs will report a single global-level equal op.
            *   Using 'stanza' with identical inputs will return all stanzas as equal.
            *   Using 'key' will ensure that two stanzas with no common keys will be reported in
                terms of key changes.  Whereas 'global' or 'stanza' would result in a single giant replace op.

    :type replace_level: str: ``global``, ``stanza``, or ``key``
    :return: a sequence of differences in tuples
    :rtype: [DiffOp]

    .. note:: The :py:class:`DiffOp` output idea was borrowed from
              :class:`SequenceMatcher` class in the :mod:`difflib`
              in the standard Python module.

    This function returns a sequence of 5 element tuples describing the
    transformation based on the detail level specified in ``replace_level``.

    Each :py:class:`DiffOp` (named tuple) takes the form:

        (tag, location, a, b)

    *tag:*

    =========   ============================================
    Value	    Meaning
    =========   ============================================
    'replace'	same element in both, but different values.
    'delete'	remove value b
    'insert'    insert value a
    'equal'	    same values in both
    =========   ============================================

    *location* is a namedtuple that can take the following forms:

    ====================== ========== ==============================================================
    Tuple form             Type       Description
    ====================== ========== ==============================================================
    `("global")`           DiffGlobal Global file level context (e.g., both files are the same)
    `("stanza", stanza)`   DiffStanza Stanzas are the same, or completely different (no shared keys)
    `("key", stanza, key)` DiffStzKey Key level change
    ====================== ========== ==============================================================

    .. versionchanged:: v0.8.8
        The ``preserve_empty`` argument was originally introduced to preserve backwards
        compatibility, but it ended up introducing new bugs.
        Additionally, no use cases were found where better to automatically discarding empty stanzas.

    .. versionchanged:: v0.8.8
        The ``allow_level0`` argument was replaced with ``replace_level``.
        Instead of using ``allow_level0=False`` use ``replace_level="stanza"``.
        At the same time a new feature was added to support ``replace_level="key"``.
        The default behavior remains the same.

    """
    # Possible alternatives:
    # https://dictdiffer.readthedocs.io/en/latest/#dictdiffer.patch

    if replace_level not in (DiffLevel.GLOBAL, DiffLevel.STANZA, DiffLevel.KEY):
        if isinstance(replace_level, str):
            replace_level = DiffLevel(replace_level)
        else:
            raise TypeError(f"Invalid value '{replace_level}' given for "
                            f"replace_level.  Choose 'global', 'stanza', or 'key'")

    delta = []

    # Level 0 - Compare entire file
    if replace_level == DiffLevel.GLOBAL:
        stanza_a, stanza_common, stanza_b = cmp_sets(list(a.keys()), list(b.keys()))
        if a == b:
            return [DiffOp(DiffVerb.EQUAL, DiffGlobal(DiffLevel.GLOBAL), a, b)]
        if not stanza_common:
            # Q:  Does this specific output make the consumer's job more difficult?
            # Nothing in common between these two files
            # Note:  Stanza renames are not detected and are out of scope.
            return [DiffOp(DiffVerb.REPLACE, DiffGlobal(DiffLevel.GLOBAL), a, b)]

    # Level 1 - Compare stanzas

    # Make sure GLOBAL stanza is output first
    all_stanzas = set(a.keys()).union(list(b.keys()))
    if GLOBAL_STANZA in all_stanzas:
        all_stanzas.remove(GLOBAL_STANZA)
        all_stanzas = [GLOBAL_STANZA] + list(all_stanzas)
    else:
        all_stanzas = list(all_stanzas)
    all_stanzas = sorted(all_stanzas)
    for stanza in all_stanzas:
        delta.extend(compare_stanzas(a.get(stanza), b.get(stanza), stanza, replace_level))
    return delta


def summarize_cfg_diffs(delta: List[DiffOp], stream: TextIO):
    """ Summarize a delta into a human-readable format.   The input ``delta`` is in the format
    produced by the compare_cfgs() function.
    """
    stanza_stats = defaultdict(set)
    key_stats = defaultdict(lambda: defaultdict(lambda: defaultdict(set)))
    c = Counter()
    for op in delta:
        c[op.tag] += 1
        if isinstance(op.location, DiffStanza):
            stanza_stats[op.tag].add(op.location.stanza)
        elif isinstance(op.location, DiffStzKey):
            key_stats[op.tag][op.location.stanza][op.location.key].add(op.location.key)

    for tag in sorted(c.keys()):
        stream.write(f"Have {c[tag]} '{tag}' operations:\n")
        for entry in sorted(stanza_stats[tag]):
            stream.write(f"\t[{_format_stanza(entry)}]\n")
        for entry in sorted(key_stats[tag]):
            stream.write(f"\t[{_format_stanza(entry)}]  {len(key_stats[tag][entry])} keys\n")
        stream.write("\n")


def is_equal(delta: List[DiffOp]) -> bool:
    """ Is the delta output show that the compared objects are identical """
    return len(delta) == 1 and delta[0].tag == DiffVerb.EQUAL


# Color mapping
_diff_color_mapping = {
    " ": ANSI_RESET,
    "+": ANSI_GREEN,
    "-": ANSI_RED,
}


def _show_diff_header(stream: TextIO,
                      files: List[Union[DiffHeader, str]],
                      diff_line: Optional[str] = None):
    headers = []

    for f in files:
        if isinstance(f, DiffHeader):
            headers.append(f)
        else:
            headers.append(DiffHeader(f))

    with TermColor(stream) as tc:
        tc.color(ANSI_YELLOW, ANSI_BOLD)
        if diff_line:
            stream.write(f"diff {diff_line} {headers[0].name} {headers[1].name}\n")
        tc.reset()
        stream.write(f"--- {headers[0]}\n")
        stream.write(f"+++ {headers[1]}\n")


def show_diff(stream: TextIO, diffs: List[DiffOp], headers=None) -> int:
    tc = TermColor(stream)

    def is_multiline(v):
        if v and "\n" in v:
            return True
        else:
            return False

    def write_key(key, value, prefix_=" "):
        if is_multiline(value):
            write_multiline_key(key, value, prefix_)
        else:
            with tc:
                tc.color(_diff_color_mapping.get(prefix_))
                if key.startswith("#-"):
                    line = value
                else:
                    line = f"{key} = {value}"
                stream.write(f"{prefix_}{line}\n")

    def write_multiline_key(key, value, prefix_=" "):
        with tc:
            lines = value.replace("\n", "\\\n").split("\n")
            tc.color(_diff_color_mapping.get(prefix_))
            stream.write(f"{prefix_}{key} = {lines.pop(0)}\n")
            for line in lines:
                stream.write(f"{prefix_}{line}\n")

    def show_value(value, stanza_, key, prefix_=""):
        with tc:
            tc.color(_diff_color_mapping.get(prefix_))
            if isinstance(value, dict):
                if stanza_ is not GLOBAL_STANZA:
                    stream.write(f"{prefix_}[{stanza_}]\n")
                for x, y in sorted(value.items()):
                    write_key(x, y, prefix_)
            else:
                write_key(key, value, prefix_)

    def show_multiline_diff(value_a, value_b, key):
        def f(v):
            r = f"{key} = {v}"
            r = r.replace("\n", "\\\n")
            return r.splitlines()

        a = f(value_a)
        b = f(value_b)
        differ = difflib.Differ()
        with tc:
            for d in differ.compare(a, b):
                # Someday add "?" highlighting.  Trick is this should change color mid-line on the
                # previous (one or two) lines.  (Google and see if somebody else solved this one already)
                # https://stackoverflow.com/questions/774316/python-difflib-highlighting-differences-inline
                tc.color(_diff_color_mapping.get(d[0], 0))
                ''' # Attempt to keep this disabled for Python 3
                # Differences in how difflib returns bytes/unicode?
                if isinstance(d, bytes):
                    d = d.decode(default_encoding)
                '''
                stream.write(d)
                tc.reset()
                stream.write("\n")

    # Global result:  no changes between files or no commonality between files
    if len(diffs) == 1 and isinstance(diffs[0].location, DiffGlobal):
        op = diffs[0]
        if op.tag == DiffVerb.EQUAL:
            return EXIT_CODE_DIFF_EQUAL
        else:
            if headers:
                _show_diff_header(stream, headers, "--ksconf -global")
            # This is the only place where a gets '-' and b gets '+'
            for (prefix, data) in [("-", op.a), ("+", op.b)]:
                for (stanza, keys) in sorted(data.items()):
                    show_value(keys, stanza, None, prefix)
            stream.flush()
            return EXIT_CODE_DIFF_NO_COMMON

    if headers:
        _show_diff_header(stream, headers, "--ksconf")

    last_stanza = None
    for op in diffs:
        if op.location.stanza != last_stanza:
            if last_stanza is not None:
                # Line break after last stanza
                stream.write("\n")
                stream.flush()
            if op.location.stanza is not GLOBAL_STANZA and not isinstance(op.location, DiffStanza):
                stream.write(f" [{op.location.stanza}]\n")
            last_stanza = op.location.stanza

        if isinstance(op.location, DiffStanza):
            if op.tag in (DIFF_OP_DELETE, DIFF_OP_REPLACE):
                show_value(op.a, op.location.stanza, None, "-")
            if op.tag in (DIFF_OP_INSERT, DIFF_OP_REPLACE):
                show_value(op.b, op.location.stanza, None, "+")
            continue

        if op.tag == DIFF_OP_INSERT:
            show_value(op.b, op.location.stanza, op.location.key, "+")
        elif op.tag == DIFF_OP_DELETE:
            show_value(op.a, op.location.stanza, op.location.key, "-")
        elif op.tag == DIFF_OP_REPLACE:
            if is_multiline(op.a) or is_multiline(op.b):
                show_multiline_diff(op.a, op.b, op.location.key)
            else:
                show_value(op.a, op.location.stanza, op.location.key, "-")
                show_value(op.b, op.location.stanza, op.location.key, "+")
        elif op.tag == DIFF_OP_EQUAL:
            show_value(op.b, op.location.stanza, op.location.key, " ")
    stream.flush()
    return EXIT_CODE_DIFF_CHANGE


def show_text_diff(stream: TextIO, a: PathLike, b: PathLike):
    _show_diff_header(stream, (a, b), "--text")
    differ = difflib.Differ()
    lines_a = open(a, "r", encoding=default_encoding).readlines()
    lines_b = open(b, "r", encoding=default_encoding).readlines()
    with TermColor(stream) as tc:
        for d in differ.compare(lines_a, lines_b):
            # Someday add "?" highlighting.  Trick is this should change color mid-line on the
            # previous (one or two) lines.  (Google and see if somebody else solved this one already)
            # https://stackoverflow.com/questions/774316/python-difflib-highlighting-differences-inline
            tc.color(_diff_color_mapping.get(d[0], 0))
            stream.write(d)
            tc.reset()


def reduce_stanza(stanza: StanzaType, keep_attrs: Sequence) -> dict:
    """ Pre-process a stanzas so that only a common set of keys will be compared.

    :param stanza: Stanzas containing attributes and values
    :type stanza: dict
    :param keep_attrs: Listing of attributes to preserve
    :type keep_attrs: (list, set, tuple, dict)
    :return: a reduced copy of ``stanza``.
    """
    return {attr: value for attr, value in stanza.items() if attr in keep_attrs}


def write_diff_as_json(delta: List[DiffOp], stream, **dump_args):
    # XXX: Eventually support reversing this to import a json "patch"; and add an apply/reverse command.
    import json
    import sys

    from ksconf.version import vcs_info, version
    record = {
        "schema_version": 1,
        "software": {
            "name": "ksconf",
            "version": [version, vcs_info],
            "command": sys.argv,
            "cwd": os.getcwd(),
        }
    }
    record["records"] = [diff_obj_json_format(op) for op in delta]
    json.dump(record, stream, **dump_args)


def diff_obj_json_format(o):
    # Q: PY3 move.  Do any of the new fancy classes above help with this? (dataclass/NamedTuple)
    if isinstance(o, DiffOp):
        o = {
            "tag": o.tag,
            "location": diff_obj_json_format(o.location),
            "a": o.a,
            "b": o.b,
        }
    elif isinstance(o, DiffGlobal):
        o = {"type": o.type}
    elif isinstance(o, DiffStanza):
        o = {"type": o.type, "stanza": o.stanza}
    elif isinstance(o, DiffStzKey):
        o = {"type": o.type, "stanza": o.stanza, "key": o.key}
    return o



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# PYTHON_ARGCOMPLETE_OK
#
# Unlike the rest of ksconf, this file should remain Python 2.7 compatible
# This is just enough to gracefully report our presence in an unsupported
# version of Python, if we happen to get installed there.  Try really hard
# to avoid confusing Python compiler errors.
""" KSCONF - Ksconf Splunk CONFig tool

Optionally supports argcomplete for commandline argument (tab) completion.

Install & register with:

     pip install argcomplete
     activate-global-python-argcomplete  (in ~/.bashrc)

"""
from __future__ import absolute_import, unicode_literals

import sys

if sys.version_info < (3, 7):   # noqa
    # Can't actually import 'consts' because it loads 'enum'; so we cheat
    EXIT_CODE_BAD_PY_VERSION = 121
    sys.stderr.write(
        "ABORT!  ksconf cannot run here because it requires Python 3.7 or later!\n\n"
        "Check to see if you have multiple version of Python installed.\n"
        "If you need Python 2.7 or 3.6 support, consider downgrading to an earlier release.\n"
        "Releases 0.9.x support these older Python versions.\n\n"
        "If you installed this with pip, try running:\n\n"
        "   pip install -U 'kintyre-splunk-conf>=0.9,<=0.10'\n")
    sys.stderr.write("\nAdditional info\n"
                     "  Python:  {}  {}\n"
                     "  Ksconf:  {}\n".format(sys.executable, sys.version_info[:3], sys.argv[0]))
    sys.exit(EXIT_CODE_BAD_PY_VERSION)

from ksconf.cli import cli

if __name__ == '__main__':  # pragma: no cover
    cli()



from __future__ import absolute_import, annotations, unicode_literals

import hashlib
import os
import re
import shutil
import tarfile
import tempfile
from functools import wraps
from os import fspath
from pathlib import Path
from typing import Optional, TextIO, Union

from ksconf.app.manifest import AppManifest
from ksconf.combine import LayerCombiner
from ksconf.conf.merge import merge_app_local, merge_conf_dicts
from ksconf.conf.parser import conf_attr_boolean, parse_conf, update_conf
from ksconf.consts import is_debug
from ksconf.hook import plugin_manager
from ksconf.util import decorator_with_opt_kwargs
from ksconf.util.file import atomic_writer
from ksconf.vc.git import git_cmd


def find_conf_in_layers(app_dir, conf, *layers):
    if not layers:
        layers = ("local", "default")
    for layer in layers:
        conf_file = os.path.join(app_dir, layer, conf)
        if os.path.isfile(conf_file):
            return conf_file


def get_merged_conf(app_dir, conf, *layers):
    if not layers:
        # Last layer wins
        layers = ("default", "local")
    files = [os.path.join(app_dir, layer, conf) for layer in layers]
    confs = [parse_conf(path) for path in files if os.path.isfile(path)]
    return merge_conf_dicts(*confs)


def normalize_directory_mtime(path):
    """ Walk a tree and update the directory modification times to match the
    newest time of the children.  This results in a more predictable behavior
    over multiple executions.
    """
    for (root, dirs, files) in os.walk(path, topdown=False):
        nodes = dirs + files
        if not nodes:
            # Empty directories are somewhat unlikely.  We'll see
            continue
        mtime = max(os.stat(os.path.join(root, n)).st_mtime for n in nodes)
        os.utime(root, (mtime, mtime))


class PackagingException(Exception):
    pass


class AppPackager:

    def __init__(self, src_path, app_name: str, output: TextIO,
                 template_variables: Optional[dict] = None,
                 predictable_mtime: bool = True):
        self.src_path = fspath(src_path)
        self.app_name = app_name
        self.build_dir: str = None
        self.app_dir: str = None
        self.output = output
        self._var_magic: AppVarMagic = None
        self._mutable: bool = None
        self._frozen_by = ""
        self.template_variables = template_variables
        self.predictable_mtime = predictable_mtime

    @decorator_with_opt_kwargs
    def require_active_context(funct, mutable=True):
        """ Decorator to mark member functions that cannot be used until the
        context manager has been activated.
        """
        @wraps(funct)
        def wrapper(self: AppPackager, *args, **kwargs):
            assert self.build_dir, "Context manager not yet active"
            if mutable:
                # Ensure make_archive() and make_manifest() are called last
                assert self._mutable, "App content already frozen. " \
                    f"Must call {funct.__name__}() before {self._frozen_by}()"
            return funct(self, *args, **kwargs)
        return wrapper

    def cleanup(self):
        # Do we need  https://stackoverflow.com/a/21263493/315892  (Windows): -- See tests/cli_helper
        shutil.rmtree(self.build_dir)
        self.build_dir = None
        self.app_dir = None
        self._mutable = None

    def expand_var(self, value: str) -> str:
        """ Expand a variable, if present

        :param str value:  String that main contain ``{{variable}}`` substitution.
        :return: Expanded value
        :rtype: str
        """
        return self._var_magic.expand(value)

    def expand_new_only(self, value: str) -> Union[str, bool]:
        """ Expand a variable but return False if no substitution occurred

        :param str value:  String that may contain ``{{variable}}`` substitution.
        :return:  Expanded value if variables were expanded, else False
        :rtype: str
        """
        new_value = self._var_magic.expand(value)
        return new_value if new_value != value else False

    @require_active_context
    def combine(self, src, filters, layer_method="dir.d", allow_symlink=False):
        combiner = LayerCombiner(follow_symlink=allow_symlink, quiet=True)
        if self.template_variables:
            combiner.context.template_variables = self.template_variables
        if layer_method == "dir.d":
            combiner.set_layer_root(src)
        elif layer_method == "disable":
            combiner.set_source_dirs([src])
        else:
            raise NotImplementedError(f"layer_method of '{layer_method}' is not supported.  "
                                      "Please use 'dir.d' or 'disable'.")
        for action, path in filters:
            combiner.add_layer_filter(action, path)
        combiner.combine(self.app_dir, hook_label="package")
        self._var_magic.meta["layers"] = combiner.layer_names_used

    @require_active_context
    def blocklist(self, patterns):
        # XXX: Rewrite explicitly blocklist '.git' dir, because '.git*' wasn't working here. :=(

        # For now we just delete files out the build directory.  Not sophisticated, but it works
        # Do we need relwalker here?  relwalk
        from fnmatch import fnmatch
        for (root, dirs, files) in os.walk(self.build_dir, topdown=True):
            for fn in files:
                path = os.path.join(root, fn)
                for pattern in patterns:
                    if ("*" in pattern and fnmatch(path, pattern)) or fn == pattern:
                        self.output.write(f"Blocked file: {path}  (pattern: {pattern})\n")
                        os.unlink(path)
                        break
            for d in list(dirs):
                path = os.path.join(root, d)
                for pattern in patterns:
                    if ("*" in pattern and fnmatch(path, pattern)) or d == pattern:
                        self.output.write(f"Blocked dir:  {path}  (pattern: {pattern})\n")
                        dirs.remove(d)
                        shutil.rmtree(path)
                        break

    @require_active_context
    def merge_local(self):
        # XXX:  Rename this "promote_local()" ?
        """
        Find everything in local, if it has a corresponding file in default, merge.
        """
        # XXX: No logging/reporting done here :-(
        merge_app_local(self.app_dir)
        # Cleanup anything remaining in local
        self.block_local(report=False)

    @require_active_context
    def block_local(self, report=True):
        local_dir = os.path.join(self.app_dir, "local")
        if os.path.isdir(local_dir):
            if report:
                self.output.write("Removing local directory.\n")
            shutil.rmtree(local_dir)
        local_meta = os.path.join(self.app_dir, "metadata", "local.meta")
        if os.path.isfile(local_meta):
            if report:
                self.output.write("Removing local.meta\n")
            os.unlink(local_meta)

    @require_active_context
    def update_app_conf(self,
                        version: Optional[str] = None,
                        build: Optional[str] = None):
        """ Update version and/or build in ``apps.conf`` """
        app_settings = [
            ("launcher", "version", version),
            ("install", "build", build),
        ]
        appconf_file = find_conf_in_layers(self.app_dir, "app.conf") or \
            os.path.join(self.app_dir, "default", "app.conf")

        self.output.write(f"Updating app.conf file:  {appconf_file}\n")
        with update_conf(appconf_file, make_missing=True) as conf:
            for (stanza, attr, value) in app_settings:
                new_value = self.expand_new_only(value)
                if value:
                    if stanza not in conf:
                        conf[stanza] = {}
                    if new_value:
                        self.output.write(f"\tUpdate app.conf:  [{stanza}] "
                                          f"{attr} = {new_value} "
                                          f"(From {value})\n")
                        value = new_value
                    else:
                        self.output.write(f"\tUpdate app.conf:  [{stanza}] "
                                          f"{attr} = {value}\n")
                    conf[stanza][attr] = value

    @require_active_context(mutable=False)
    def check(self):
        """ Run safety checks prior to building archive:

        1.  Set app name based on app.conf [package] id, if set.  Otherwise confirm that the package
            id and top-level folder names align.
        2.  Check for files or directories starting with ``.``, makes AppInspect very grumpy!
        """
        # Should this force a freeze?  (Disable if this breaks anything...)
        self.freeze("check")
        app_conf = get_merged_conf(self.app_dir, "app.conf")
        try:
            package_id = app_conf["package"]["id"]
            target_splunkbase = conf_attr_boolean(app_conf["package"]
                                                  .get("check_for_updates", "false"))
        except KeyError:
            self.output.write("Skipped folder and package id check due to missing app.conf entry\n")
            package_id = None

        if package_id:
            if not self.app_name or self.app_name == ".":
                self.output.write("Set app name from app.conf:  "
                                  f"{self.app_name}\n")
                self.app_name = package_id
            elif package_id != self.app_name:
                self.output.write(f"Top-level folder does not match the "
                                  f"package id:  folder: {self.app_name} "
                                  f"package id: {package_id}\n")
                if target_splunkbase:
                    raise PackagingException("Aborting build due to app name and package id "
                                             "discrepancy for public app:  "
                                             f"{package_id} != {self.app_name}")

        for root, dirs, files in os.walk(self.app_dir):
            for items, t in [(dirs, "directory"), (files, "file")]:
                for name in items:
                    if name[0] == ".":
                        self.output.write(f"Found hidden {t}:  {root}/{name}\n")

    @require_active_context(mutable=False)
    def make_archive(self, filename: str, temp_suffix: str = ".tmp") -> str:
        """ Create a compressed tarball of the build directory.
        """
        self.freeze("make_archive")
        # if os.path.isfile(filename):
        #    raise ValueError(f"Destination file already exists:  {filename}")
        app_name = self.expand_var(self.app_name)
        if app_name != self.app_name:
            self.output.write(f"Expanding template {self.app_name} to final "
                              f"app name: {app_name}\n")

        new_filename = self.expand_new_only(filename)
        if new_filename:
            self.output.write(f"Creating archive:  {new_filename}  (Expanded "
                              f"from '{os.path.basename(filename)}'\n")
            filename = new_filename
        else:
            self.output.write(f"Creating archive:  {filename}\n")

        if self.predictable_mtime:
            normalize_directory_mtime(self.app_dir)

        with atomic_writer(filename, temp_name=temp_suffix) as tmp_filename:
            with tarfile.open(tmp_filename, mode="w:gz") as spl:
                spl.add(self.app_dir, arcname=self.app_name)
        return filename

    @require_active_context(mutable=False)
    def make_manifest(self, calculate_hash=True) -> AppManifest:
        """
        Create a manifest of the app's contents.
        """
        self.freeze("make_manifest")
        app_name = self.expand_var(self.app_name)
        manifest = AppManifest.from_filesystem(self.app_dir, name=app_name,
                                               calculate_hash=calculate_hash)
        manifest.source = self.src_path
        return manifest

    def freeze(self, caller_name):
        """
        Initiate a content freeze by restricting mutable methods.
        The "package_pre_archive" hook is invoked before freeze operation.
        Such hooks may choose to mutate the filesystem at ``app_dir``, the only
        assumption is that all work is done before the hook returns.

        Freeze can be safely called multiple times. ``caller_name`` is simply a
        label used in an exception message if the programmer screwed up.
        """
        if self._mutable:
            plugin_manager.hook.package_pre_archive(app_dir=Path(self.app_dir),
                                                    app_name=self.expand_var(self.app_name))
            # Forbid any additional internal changes to the package
            self._mutable = False
            self._frozen_by = caller_name

    def __enter__(self) -> AppPackager:
        self.build_dir = tempfile.mkdtemp("-ksconf-package-build")
        self._mutable = True
        if self.app_name == "." or "{{" in self.app_name:
            # Use a placeholder app name, otherwise build_dir == app_dir
            self.app_dir = os.path.join(self.build_dir, "app")
        else:
            self.app_dir = os.path.join(self.build_dir, self.app_name)
        self._var_magic = AppVarMagic(self.src_path, self.app_dir)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cleanup()


class AppVarMagicException(KeyError):
    pass


class AppVarMagic:
    """ A lazy loading dict-like object to fetch things like app version and such on demand. """

    def __init__(self, src_dir, build_dir, meta=None):
        self._cache = {}
        self.src_dir = src_dir
        self.build_dir = build_dir
        self.meta = meta or {}

    def expand(self, value: str) -> str:
        """ A simple Jinja2 like ``{{VAR}}`` substitution mechanism. """
        def replace(match_obj):
            var = match_obj.group(1)
            return self[var]
        if value:
            value = str(value)
            return re.sub(r"\{\{\s*([\w_]+)\s*\}\}", replace, value)
        return value

    def git_single_line(self, *args):
        out = git_cmd(args, cwd=self.src_dir)
        if out.returncode != 0:
            return f"git-errorcode-{out.returncode}"
        return out.stdout.strip()

    # START Variable fetching functions.  Be sure to add a docstring

    def get_version(self):
        """ Splunk app version fetched from app.conf """
        app_conf = get_merged_conf(self.build_dir, "app.conf")
        try:
            return app_conf["launcher"]["version"]
        except KeyError as e:
            raise AppVarMagicException(e)

    def get_build(self):
        """ Splunk app build fetched from app.conf """
        app_conf = get_merged_conf(self.build_dir, "app.conf")
        try:
            return app_conf["install"]["build"]
        except KeyError as e:
            raise AppVarMagicException(e)

    def get_app_id(self):
        """ Splunk app package id from app.conf """
        app_conf = get_merged_conf(self.build_dir, "app.conf")
        try:
            return app_conf["package"]["id"]
        except KeyError as e:
            raise AppVarMagicException(e)

    def get_git_tag(self):
        """ Git version tag using the ``git describe --tags`` command """
        tag = self.git_single_line("describe", "--tags", "--always", "--dirty")
        return re.sub(r'^(v|release|version)-?', "", tag)

    def get_git_last_rev(self):
        """ Git abbreviated rev of the last change of the app.  This may not be the same as HEAD. """
        return self.git_single_line("log", "-n1", "--pretty=format:%h", "--", ".")

    def get_git_head(self):
        """ Git HEAD rev abbreviated """
        return self.git_single_line("rev-parse", "--short", "HEAD")

    def get_layers_list(self):
        """ List of ksconf layers used. """
        layers = sorted(self.meta.get("layers"))
        if layers:
            return f'__{"__".join(layers)}__'
        else:
            return ""

    def get_layers_hash(self):
        """ Build a unique hash representing the combination of ksconf layers used. """
        DIGITS = 16
        layers_string = self["layers_list"]
        if layers_string:
            h = hashlib.sha256(layers_string.encode("utf-8"))
            return h.hexdigest()[:DIGITS]
        else:
            return "0" * DIGITS

    # END Variable fetching functions.

    def list_vars(self):
        """ Return a list of (variable, description) available in this class. """
        for name in dir(self):
            if name.startswith("get_"):
                var = name[4:]
                doc = getattr(self, name).__doc__.strip()
                yield (var, doc)

    def __getitem__(self, item):
        if item not in self._cache:
            self._cache[item] = self._get_expanded_var(item)
        return self._cache[item]

    def _get_expanded_var(self, item):
        get_funct_name = "get_" + item
        if hasattr(self, get_funct_name):
            try:
                funct = getattr(self, get_funct_name)
                return funct()
            except AppVarMagicException as e:
                if is_debug():
                    raise e
                return f"VAR-{item}-ERROR"
        else:
            raise KeyError(item)



""" SUBCOMMAND:  ``ksconf minimize --target=<CONF> <CONF> [ <CONF-n> ... ]``

Usage example:

.. code-block:: sh

    ksconf minimize --target=local/inputs.conf default/inputs.conf

Example workflow:
   1. ``cp default/props.conf local/props.conf``
   2. ``vi local/props.conf`` (edit JUST the lines you want to change)
   3. ``ksconf minimize --target=local/props.conf default/props.conf``
      (You could take this a step further by appending "$SPLUNK_HOME/system/default/props.conf"
      and removing any SHOULD_LINEMERGE = true entries (for example)

"""
from __future__ import absolute_import, unicode_literals

from ksconf.command import ConfFileType, KsconfCmd, dedent
from ksconf.conf.delta import (DIFF_OP_DELETE, DIFF_OP_EQUAL, DIFF_OP_INSERT,
                               DIFF_OP_REPLACE, DiffStanza, compare_cfgs,
                               show_diff)
from ksconf.conf.merge import merge_conf_dicts
from ksconf.conf.parser import (GLOBAL_STANZA, PARSECONF_LOOSE,
                                PARSECONF_STRICT, _drop_stanza_comments)
from ksconf.consts import EXIT_CODE_MISSING_ARG
from ksconf.filter import create_filtered_list
from ksconf.util.completers import conf_files_completer


def explode_default_stanza(conf, default_stanza=None):
    """ Take the GLOBAL stanza, (aka [default]) and apply it's settings underneath ALL other
    stanzas.  This is mostly only useful in minimizing and other comparison operations. """
    if default_stanza is None:
        default_stanza = conf.get(GLOBAL_STANZA, conf.get("default"))
        if not default_stanza:
            return conf
    default_stanza = _drop_stanza_comments(default_stanza)
    n = {}
    for (stanza, content) in conf.items():
        new_content = dict(default_stanza)
        new_content.update(content)
        n[stanza] = new_content
    return n


class MinimizeCmd(KsconfCmd):
    help = "Minimize the target file by removing entries duplicated in the default conf(s)"
    description = dedent("""\
    Minimize a conf file by removing any duplicated default settings.

    Reduce a local conf file to only your intended changes without manually tracking
    which entries you've edited.  Minimizing local conf files makes your local
    customizations easier to read and often results in cleaner upgrades.
    """)
    maturity = "beta"

    def register_args(self, parser):
        parser.add_argument("conf", metavar="CONF", nargs="+",
                            type=ConfFileType("r", "load", parse_profile=PARSECONF_LOOSE),
                            help="The default configuration file(s) used to determine what base settings are. "
                            "The base settings determine what is unnecessary to repeat in target file."
                            ).completer = conf_files_completer
        parser.add_argument("--target", "-t", metavar="TARGET",
                            type=ConfFileType("r+", "load", parse_profile=PARSECONF_STRICT),
                            help="The local file that you wish to remove duplicate settings from.  "
                            "This file will be read from and then replaced with a minimized version."
                            ).completer = conf_files_completer
        grp1 = parser.add_mutually_exclusive_group()
        grp1.add_argument("--dry-run", "-D", default=False, action="store_true",
                          help="Enable dry-run mode.  "
                          "Instead of writing and minimizing the TARGET file, preview what would be removed "
                          "as a 'diff'.")
        grp1.add_argument("--output",
                          type=ConfFileType("w", "none", parse_profile=PARSECONF_STRICT),
                          default=None,
                          help="Write the minimized output to a separate file instead of updating TARGET."
                          ).completer = conf_files_completer
        parser.add_argument("--explode-default", "-E", default=False, action="store_true",
                            help="Enable minimization across stanzas for special use-cases.  "
                            "Helpful when dealing with stanzas downloaded from a REST endpoint or "
                            "``btool list`` output.")
        parser.add_argument("-k", "--preserve-key", action="append", default=[],
                            help="Specify attributes that should always be kept.")

    def run(self, args):
        if not args.target:
            self.stderr.write("You must specify the target file with '--target'.\n")
            return EXIT_CODE_MISSING_ARG
        if args.explode_default:
            # Is this the SAME as exploding the defaults AFTER the merge?;
            # I think NOT.  Needs testing
            cfgs = [explode_default_stanza(conf.data) for conf in args.conf]
        else:
            cfgs = [conf.data for conf in args.conf]
        # Merge all config files:
        default_cfg = merge_conf_dicts(*cfgs)
        del cfgs
        local_cfg = args.target.data
        orig_cfg = dict(args.target.data)

        if args.explode_default:
            # Make a skeleton default dict; at the highest level, that ensure that all default
            default_stanza = default_cfg.get(GLOBAL_STANZA, default_cfg.get("default"))
            skeleton_default = dict([(k, {}) for k in args.target.data])
            skeleton_default = explode_default_stanza(skeleton_default, default_stanza)
            default_cfg = merge_conf_dicts(skeleton_default, default_cfg)

            local_cfg = explode_default_stanza(local_cfg)
            local_cfg = explode_default_stanza(local_cfg, default_stanza)

        minz_cfg = dict(local_cfg)

        # This may be a bit too simplistic.  Weird interplay may exists between a [default] stanza
        # and local [Upstream] stanza line up, but [Upstream] in our default file does not.
        # XXX:  Add a unit test!

        diffs = compare_cfgs(default_cfg, local_cfg, replace_level="stanza")

        preserve_attributes = create_filtered_list("splunk", default=False)
        preserve_attributes.feedall(args.preserve_key)

        for op in diffs:
            if op.tag == DIFF_OP_DELETE:
                # This is normal.  Don't expect all default content to be mirrored into local
                continue
            elif op.tag == DIFF_OP_EQUAL:
                if isinstance(op.location, DiffStanza):
                    del minz_cfg[op.location.stanza]
                else:
                    # Todo: Only preserve keys for stanzas where at least 1 key has been modified
                    if preserve_attributes.match(op.location.key):
                        '''
                        self.stderr.write("Skipping key [PRESERVED]  [{0}] key={1} value={2!r}\n"
                                     "".format(op.location.stanza, op.location.key, op.a))
                        '''
                        continue  # pragma: no cover  (peephole optimization)
                    del minz_cfg[op.location.stanza][op.location.key]
                    # If that was the last remaining key in the stanza, delete the entire stanza
                    if not _drop_stanza_comments(minz_cfg[op.location.stanza]):
                        del minz_cfg[op.location.stanza]
            elif op.tag == DIFF_OP_INSERT:
                '''
                self.stderr.write("Keeping local change:  <{0}> {1!r}\n-{2!r}\n+{3!r}\n\n\n".format(
                    op.tag, op.location, op.b, op.a))
                '''
                continue
            elif op.tag == DIFF_OP_REPLACE:
                '''
                self.stderr.write("Keep change:  <{0}> {1!r}\n-{2!r}\n+{3!r}\n\n\n".format(
                    op.tag, op.location, op.b, op.a))
                '''
                continue

        if args.dry_run:
            if args.explode_default:
                rc = show_diff(self.stdout, compare_cfgs(orig_cfg, minz_cfg),
                               headers=(args.target.name, args.target.name + "-new"))
            else:
                rc = show_diff(self.stdout, compare_cfgs(local_cfg, default_cfg),
                               headers=(args.target.name, args.target.name + "-new"))
            return rc

        if args.output:
            args.output.dump(minz_cfg)
        else:
            args.target.dump(minz_cfg)
            '''
            # Makes it really hard to test if you keep overwriting the source file...
            print "Writing config to STDOUT...."
            write_conf(self.stdout, minz_cfg)
            '''
        # Todo:  return ?  Should only be updating target if there's a change; RC should reflect this



"""
SUBCOMMAND:  ``ksconf merge --target=<TARGET_CONF> <CONF> [ <CONF-n> ... ]``

Usage example:

.. code-block:: sh

    ksconf merge --target=master-props.conf /opt/splunk/etc/apps/*TA*/{default,local}/props.conf

"""
from __future__ import absolute_import, unicode_literals

import os

from ksconf.command import ConfFileProxy, ConfFileType, KsconfCmd, dedent
from ksconf.conf.merge import merge_conf_files
from ksconf.conf.parser import PARSECONF_MID, PARSECONF_STRICT
from ksconf.consts import EXIT_CODE_BAD_ARGS, EXIT_CODE_SUCCESS
from ksconf.util.completers import conf_files_completer


class MergeCmd(KsconfCmd):
    help = "Merge two or more .conf files"
    description = dedent("""\
    Merge two or more .conf files into a single combined .conf file.
    This is similar to the way that Splunk logically combines the ``default`` and ``local``
    folders at runtime.
    """)
    maturity = "stable"

    def register_args(self, parser):
        parser.add_argument("conf", nargs="+",
                            help="The source configuration file(s) to collect settings from."
                            ).completer = conf_files_completer
        parser.add_argument("--target", "-t",
                            type=ConfFileType("r+", "none", parse_profile=PARSECONF_STRICT),
                            default=ConfFileProxy("<stdout>", "w", self.stdout), help=dedent("""\
            Destination file for merged configurations.
            If not provided, the merged conf is written to standard output.""")
                            ).completer = conf_files_completer

        # This is helpful when writing bash expressions like MyApp/{default,local}/props.conf;
        # when either default or local may not be present.
        parser.add_argument("--ignore-missing", "-s", default=False, action="store_true",
                            help="Silently ignore any missing CONF files.")

        parser.add_argument("--in-place", "-i", default=False, action="store_true",
                            help=dedent("""
            Enable in-place update mode.  When selected, the TARGET file will also be considered as
            the base of the merge operation.  All CONF files will be merged with TARGET.
            When disabled, any existing content within TARGET is ignored and overwritten."""))

        parser.add_argument("--dry-run", "-D", default=False, action="store_true", help=dedent("""\
            Enable dry-run mode.
            Instead of writing to TARGET, preview changes in 'diff' format.
            If TARGET doesn't exist, then show the merged file."""))
        parser.add_argument("--banner", "-b", default="", help=dedent("""\
            A banner or warning comment added to the top of the TARGET file.
            Used to discourage Splunk admins from editing an auto-generated file."""))

    def pre_run(self, args):
        if args.in_place:
            if args.target.name == "<stdout>":
                self.stderr.write("In-place mode require '--target=FILE'.\n")
                return EXIT_CODE_BAD_ARGS
            # Insert target as the first source CONF file to parse
            args.conf.insert(0, args.target.name)

    def run(self, args):
        ''' Merge multiple configuration files into one '''
        self.parse_profile = PARSECONF_MID

        if args.ignore_missing:
            cfgs = [self.parse_conf(c) for c in args.conf if os.path.isfile(c) or c == "-"]
        else:
            cfgs = [self.parse_conf(conf) for conf in args.conf]

        merge_conf_files(args.target, cfgs, dry_run=args.dry_run,
                         banner_comment=args.banner)
        return EXIT_CODE_SUCCESS
